import asyncio
from bisect import bisect_left
from dataclasses import dataclass
import datetime as dt
from decimal import Decimal
from httpx import ConnectError
from typing import Dict, List
import certifi
import httpx
import ssl
from optrabot.exceptions.orderexceptions import PlaceOrderException, PrepareOrderException
from optrabot.optionhelper import OptionHelper
from optrabot.broker.optionpricedata import OptionStrikeData, OptionStrikePriceData
from optrabot.broker.brokerconnector import BrokerConnector, SymbolData
from pydantic import ValidationError
from loguru import logger
from datetime import date, timedelta
import re
from tastytrade import Account, AlertStreamer, DXLinkStreamer, Session
from optrabot.models import Account as ModelAccount
from tastytrade.instruments import NestedOptionChain, NestedOptionChainExpiration, Strike
from tastytrade.utils import TastytradeError
from tastytrade.dxfeed import Greeks, Quote, Candle
from tastytrade.instruments import Option, OptionType
from tastytrade.order import NewOrder, NewComplexOrder, OrderTimeInForce, OrderType, OrderAction, PlacedOrder, OrderStatus
import optrabot.config as optrabotcfg
from optrabot.broker.order import OptionRight, Order as GenericOrder, OrderAction as GenericOrderAction, Leg as GenericOrderLeg, OrderStatus as GenericOrderStatus, PriceEffect
from optrabot.tradetemplate.templatefactory import Template
import optrabot.symbolinfo as symbolInfo
from optrabot.managedtrade import ManagedTrade
from websockets import ConnectionClosedOK, ConnectionClosedError

@dataclass
class TastySymbolData(SymbolData):
	def __init__(self) -> None:
		super().__init__()
		self.tastySymbol: str = None
		self.chain: NestedOptionChain = None

class TastytradeConnector(BrokerConnector):
	task_listen_quotes: asyncio.Task = None

	def __init__(self) -> None:
		super().__init__()
		self._username = ''
		self._password = ''
		self._sandbox = False
		self._initialize()
		self.id = 'TASTY'
		self.broker = 'TASTY'
		self._orders: List[GenericOrder] = []
		self._replacedOrders: List[PlacedOrder] = []
		self._processed_fills: set = set()  # OTB-267: Track processed fills to avoid duplicate transactions
		self._is_disconnecting = False
		self._session = None
		self._streamer: DXLinkStreamer = None
		self._alert_streamer: AlertStreamer = None
		self._quote_symbols = []
		self._candle_symbols = []
		self._greeks_symbols = []
		self._tasty_accounts: List[Account] = []
		self._symbolReverseLookup: Dict[str, str] = {}		# maps tastytrade symbol to generic symbol

		self.task_listen_quotes = None
		self.task_listen_accounts = None
		self.task_listen_greeks = None
		self.task_listen_candle = None

	def _initialize(self):
		"""
		Initialize the Tastytrade connector from the configuration
		"""
		if not optrabotcfg.appConfig:
			return
		
		config :optrabotcfg.Config = optrabotcfg.appConfig
		try:
			config.get('tastytrade')
		except KeyError as keyErr:
			logger.debug('No Tastytrade connection configured')
			return
		
		try:
			self._username = config.get('tastytrade.username')
		except KeyError as keyErr:
			logger.error('Tastytrade username not configured')
			return
		try:
			self._password = config.get('tastytrade.password')
		except KeyError as keyErr:
			logger.error('Tastytrade password not configured')
			return
		
		try:
			self._sandbox = config.get('tastytrade.sandbox')
		except KeyError as keyErr:
			pass
		self._initialized = True

	async def cancel_order(self, order: GenericOrder):
		""" 
		Cancels the given order
		"""
		await super().cancel_order(order)
		tasty_order: PlacedOrder = order.brokerSpecific['order']
		account: Account = order.brokerSpecific['account']
		logger.debug(f'Cancelling order {tasty_order.id}')
		account.delete_order(self._session, tasty_order.id)

	async def connect(self):
		await super().connect()
		self._is_disconnecting = False
		try:
			self._session = Session(self._username, self._password, is_test=self._sandbox)
			await self.set_trading_enabled(True, "Broker connected")
			
			# OTB-xxx: Verify session is ready by attempting to retrieve accounts
			# This prevents race conditions where connectedEvent is emitted before
			# the session is fully initialized and ready to handle API calls.
			# If successful, accounts are cached for subsequent getAccounts() calls.
			max_retries = 3
			session_ready = False
			for attempt in range(max_retries):
				try:
					self._tasty_accounts = Account.get(self._session)
					if len(self._tasty_accounts) > 0:
						logger.debug(f"Session verified - {len(self._tasty_accounts)} account(s) accessible")
						session_ready = True
						break
					else:
						logger.debug(f"Session not ready - no accounts found (attempt {attempt + 1}/{max_retries})")
				except Exception as e:
					logger.debug(f"Session verification failed (attempt {attempt + 1}/{max_retries}): {e}")
				
				if attempt < max_retries - 1:
					await asyncio.sleep(0.2)  # Brief delay between retries
			
			if not session_ready:
				logger.warning("Session created but verification failed after 3 attempts - proceeding with caution")
			
			self._emitConnectedEvent()
		except TastytradeError as tastyErr:
			logger.error('Failed to connect to Tastytrade: {}', tastyErr)
			self._emitConnectFailedEvent()
		except ConnectError as connect_error:
			logger.debug(f'Failed to connect to Tastytrade - network connection error')
			self._emitConnectFailedEvent()

	async def _disconnect_internal(self):
		"""
		Perform the operations for disconnecting from Tastytrade
		"""
		# Note: _is_disconnecting should already be set to True before calling this method
		logger.debug('Starting _disconnect_internal')
		await self.set_trading_enabled(False, "Broker disconnected")

		# FIRST: Close streamers gracefully to stop their internal _connect tasks
		# This prevents the RecursionError from nested task cancellation
		if self._streamer:
			try:
				logger.debug('Closing DXLink streamer')
				await asyncio.wait_for(self._streamer.close(), timeout=2.0)
				logger.debug('DXLink streamer closed')
			except asyncio.TimeoutError:
				logger.warning('Timeout closing DXLink streamer')
			except Exception as e:
				logger.debug(f'Error closing DXLink streamer: {e}')
		
		if self._alert_streamer:
			try:
				logger.debug('Closing Alert streamer')
				await asyncio.wait_for(self._alert_streamer.close(), timeout=2.0)
				logger.debug('Alert streamer closed')
			except asyncio.TimeoutError:
				logger.warning('Timeout closing Alert streamer')
			except Exception as e:
				logger.debug(f'Error closing Alert streamer: {e}')

		# SECOND: Now cancel our listener tasks (which should already be stopping)
		tasks_to_cancel = [
			task for task in [
				self.task_listen_quotes,
				self.task_listen_greeks,
				self.task_listen_candle,
				self.task_listen_accounts
			] if task and not task.done()
		]

		if tasks_to_cancel:
			logger.debug(f'Cancelling {len(tasks_to_cancel)} listener tasks')
			for task in tasks_to_cancel:
				task.cancel()
			
			# Give tasks a moment to process cancellation
			await asyncio.sleep(0.1)

		# THIRD: Clear streamer references
		logger.debug('Clearing streamer references')
		self._streamer = None
		self._alert_streamer = None
		logger.debug('Streamer references cleared')

		# FOURTH: Let event loop clean up cancelled tasks
		if tasks_to_cancel:
			logger.debug(f'Listener tasks cancelled - will be cleaned up by event loop')
		
		# Give event loop time to process cancellations
		await asyncio.sleep(0.1)

		# Clear task references
		self.task_listen_quotes = None
		self.task_listen_greeks = None
		self.task_listen_candle = None
		self.task_listen_accounts = None

		# Close session
		if self._session:
			try:
				self._session.destroy()
			except ConnectError:
				pass
			self._session = None
		
		# Clear cached account data
		self._tasty_accounts = []
		
		self._emitDisconnectedEvent()

	async def disconnect(self):
		"""
		Disconnect from Tastytrade broker.
		Sets _is_disconnecting flag BEFORE starting disconnect operations to prevent
		reconnect attempts from disconnect callbacks.
		"""
		await super().disconnect()
		# Set flag BEFORE calling _disconnect_internal to prevent reconnect attempts
		self._is_disconnecting = True
		await self._disconnect_internal()

	def getAccounts(self) -> List[ModelAccount]:
		"""
		Returns the Tastytrade accounts and starts the account update task for listening to updates on orders.
		Uses cached account data from connect() if available.
		"""
		if len(self._managedAccounts) == 0 and self.isConnected():
			# If accounts were not already loaded in connect(), load them now
			if len(self._tasty_accounts) == 0:
				self._tasty_accounts = Account.get(self._session)
			
			for tastyAccount in self._tasty_accounts:
				account = ModelAccount(id = tastyAccount.account_number, name = tastyAccount.nickname, broker = self.broker, pdt = not tastyAccount.day_trader_status)
				self._managedAccounts.append(account)

			# Store the task reference so we can cancel it during disconnect
			self.task_listen_accounts = asyncio.create_task(self._request_account_updates())
		return self._managedAccounts
	
	def isConnected(self) -> bool:
		if self._session != None:
			return True
		
	async def prepareOrder(self, order: GenericOrder, need_valid_price_data: bool = True) -> None:
		"""
		Prepares the given order for execution

		Args:
			order: The order to prepare
			need_valid_price_data: If True, validates price data availability and freshness.
								   If False, creates order legs without price validation (e.g., for trade recovery)

		Raises:
			PrepareOrderException: If the order preparation fails with a specific reason.
		"""
		symbolData = self._symbolData[order.symbol]
		comboLegs: list[GenericOrderLeg] = []
		for leg in order.legs:
			optionInstrument: Option = None
			
			if need_valid_price_data:
				# Normal order flow: use price data with validation
				try:
					optionPriceData = symbolData.optionPriceData[leg.expiration]
				except KeyError as keyErr:
					raise PrepareOrderException(f'No option price data for expiration date {leg.expiration} available!', order)

				try:
					priceData: OptionStrikePriceData = optionPriceData.strikeData[leg.strike]
					if not priceData.is_outdated():
						if leg.right == OptionRight.CALL:
							leg.askPrice = float(priceData.callAsk)
							if leg.askPrice == None:
								leg.askPrice = 0
							leg.bidPrice = float(priceData.callBid)
							if leg.bidPrice == None:
								leg.bidPrice = 0
							optionInstrument = Option.get(self._session, priceData.brokerSpecific['call_option'])
						elif leg.right == OptionRight.PUT:
							leg.askPrice = float(priceData.putAsk)
							if leg.askPrice == None:
								leg.askPrice = 0
							leg.bidPrice = float(priceData.putBid)
							if leg.bidPrice == None:
								leg.bidPrice = 0
							optionInstrument = Option.get(self._session, priceData.brokerSpecific['put_option'])
					else:
						raise PrepareOrderException(f'Price data for strike {leg.strike} is outdated or not available!', order)

				except KeyError as keyErr:
					# No data for strike available
					raise PrepareOrderException(f'No option price data for strike {leg.strike} available!', order)
				except Exception as excp:
					raise PrepareOrderException(f'Error preparing order: {excp}', order)
			else:
				# Trade recovery flow: create option instrument without price data validation
				try:
					# Use the cached option chain from symbolData
					if symbolData.chain is None:
						raise PrepareOrderException(
							f'No option chain available for {order.symbol} - symbol data may not be initialized', 
							order
						)
					
					# Find the specific option in the cached chain
					for expiration in symbolData.chain.expirations:
						if expiration.expiration_date == leg.expiration:
							for strike in expiration.strikes:
								if abs(float(strike.strike_price) - leg.strike) < 0.01:  # Float comparison tolerance
									if leg.right == OptionRight.CALL and strike.call is not None:
										optionInstrument = Option.get(self._session, strike.call)
										leg.askPrice = 0
										leg.bidPrice = 0
										break
									elif leg.right == OptionRight.PUT and strike.put is not None:
										optionInstrument = Option.get(self._session, strike.put)
										leg.askPrice = 0
										leg.bidPrice = 0
										break
							if optionInstrument:
								break
					
					if optionInstrument is None:
						raise PrepareOrderException(
							f'Could not find option instrument for {order.symbol} {leg.expiration} {leg.strike} {leg.right}', 
							order
						)
				except PrepareOrderException:
					raise
				except Exception as e:
					raise PrepareOrderException(f'Error creating option instrument for trade recovery: {e}', order)
			
			# Build the leg for the tasty trade order
			mapped_action = self._mappedOrderAction(order.action, leg.action)
			logger.debug(f'Mapping order.action={order.action} leg.action={leg.action} -> mapped_action={mapped_action}')
			comboLeg = optionInstrument.build_leg(quantity=Decimal(leg.quantity * order.quantity), action=mapped_action)
			comboLegs.append(comboLeg)

		order.brokerSpecific['comboLegs'] = comboLegs
		order.determine_price_effect()
		return True
	
	def _transform_generic_order(self, generic_order: GenericOrder) -> NewOrder:
		"""
		Transforms the given generic order to a tastytrade order
		"""
		new_order_legs = generic_order.brokerSpecific['comboLegs']
		price = abs(generic_order.price)
		tasty_price = Decimal(price * -1 if generic_order.price_effect == PriceEffect.DEBIT else price)
		rounded_tasty_price: Decimal = round(tasty_price, 2)
		new_order = None
		if generic_order.type == OrderType.LIMIT:
			new_order = NewOrder(
				time_in_force=OrderTimeInForce.DAY,
				order_type=generic_order.type,	
				legs=new_order_legs,
				price=rounded_tasty_price
			)
		elif generic_order.type == OrderType.STOP:
			if len(generic_order.legs) == 1:
				new_order = NewOrder(
					time_in_force=OrderTimeInForce.DAY,
					order_type=generic_order.type,	
					legs=new_order_legs,
					stop_trigger=rounded_tasty_price
				)
			else:
				# Stop Orders with multiple legs must be Stop Limit Orders
				calculated_limit_price = OptionHelper.roundToTickSize(round(rounded_tasty_price * Decimal('1.25'), 2))  # Round to the nearest tick size
				tasty_limit_price: Decimal = Decimal(calculated_limit_price)
				
				new_order = NewOrder(
					time_in_force=OrderTimeInForce.DAY,
					order_type=OrderType.STOP_LIMIT,
					legs=new_order_legs,
					stop_trigger=Decimal(abs(rounded_tasty_price)),
					price=tasty_limit_price
				)
		return new_order
	
	async def placeOrder(self, managed_trade: ManagedTrade, order: GenericOrder, parent_order: GenericOrder = None) -> None:
		""" 
		Places the given order for a managed account via the broker connection.
		
		Raises:
			PlaceOrderException: If the order placement fails with a specific reason.
		"""
		account = Account.get(self._session, managed_trade.template.account)
		newOrder = self._transform_generic_order(order)
		try:
			response = account.place_order(self._session, newOrder, dry_run=False)
			#placedComplexOrders = account.get_live_complex_orders(session=self._session)
			#placedOrders = account.get_live_orders(session=self._session)
			#for order in placedOrders:
			#	logger.debug(f'Live order: {order.id} underlying: {order.underlying_symbol}')
			#	#account.delete_order(session=self._session, order_id=order.id)
			logger.debug(f'Response of place Order: {response}')
			if response.errors:
				for errorMessage in response.errors:
					raise PlaceOrderException(errorMessage, order=order)
			if response.warnings:
				for warningMessage in response.warnings:
					logger.warning(f'Warning placing order: {warningMessage}')
			
			order.brokerSpecific['order'] = response.order
			order.brokerSpecific['account'] = account
			self._orders.append(order)
			logger.debug(f'Order {response.order.id} placed successfully')
		except TastytradeError as tastyErr:
			raise PlaceOrderException(f'{tastyErr}', order=order)
		except ValidationError as valErr:
			logger.error(f'Validation error placing order: {valErr}')
			err = valErr.errors()[0]
			logger.error(err)
		except Exception as exc:
			raise PlaceOrderException(f'{exc}', order=order)

	async def place_complex_order(self, take_profit_order: GenericOrder, stop_loss_order: GenericOrder, template: Template) -> bool:
		"""
		Places the Take Profit and Stop Loss Order as complex order
		"""
		account = Account.get(self._session, template.account)
		new_order_tp = self._transform_generic_order(take_profit_order)
		new_order_sl = self._transform_generic_order(stop_loss_order)
		oco_order = NewComplexOrder( orders=[ new_order_tp, new_order_sl ] )
		try:
			dry_run = False
			response = account.place_complex_order(self._session, oco_order, dry_run=dry_run)
			logger.debug(f'Response of place_complex_order: {response}')
			if response.errors:
				for errorMessage in response.errors:
					logger.error(f'Error placing order: {errorMessage}')
					return False
			if response.warnings:
				for warningMessage in response.warnings:
					logger.warning(f'Warning placing order: {warningMessage}')
			
			try:
				take_profit_order.brokerSpecific['order'] = response.complex_order.orders[0]
				take_profit_order.brokerSpecific['account'] = account
				self._orders.append(take_profit_order)
			except IndexError as indexErr:
				logger.error(f'Error complex order. Take Profit order not found in complex order response!')
				return False

			try: 
				stop_loss_order.brokerSpecific['order'] = response.complex_order.orders[1]
				stop_loss_order.brokerSpecific['account'] = account
				self._orders.append(stop_loss_order)
			except IndexError as indexErr:
				logger.error(f'Error complex order. Stop Loss order not found in complex order response!')
				return False

			logger.debug(f'Complex Order {response.complex_order.id} placed successfully')
			return True
		except TastytradeError as tastyErr:
			logger.error(f'Error placing order: {tastyErr}')
		except ValidationError as valErr:
			logger.error(f'Validation error placing order: {valErr}')
			err = valErr.errors()[0]
			logger.error(err)
			#logger.error(repr(valErr.errors()[0]['type']))
		except Exception as exc:
			logger.error(f'Unexpected exception placing order: {exc}')
			
		return False		
		
	async def adjustOrder(self, managed_trade: ManagedTrade, order: GenericOrder, price: float) -> bool:
		""" 
		Adjusts the given order with the given new price
		"""
		if order.status == GenericOrderStatus.FILLED:
			logger.info('Order {} is already filled. Adjustment not required.', order)
			return True
		try:
			tasty_order: PlacedOrder = order.brokerSpecific['order']
		except KeyError as keyErr:
			logger.error(f'Order {order.id} not prepared for adjustment. Cannot adjust order.')
			return False
		
		account: Account = order.brokerSpecific['account']
		logger.debug(f'Adjusting order {tasty_order.id} to price {price}')

		order.price = price
		replacement_order = self._transform_generic_order(order)

		#new_order_legs = order.brokerSpecific['comboLegs']
		#new_price = abs(price)
		#tasty_price = Decimal(new_price * -1 if order.price_effect == PriceEffect.DEBIT else new_price)

		# if order.type == OrderType.LIMIT:
		# 	replacement_order = NewOrder(
		# 		time_in_force=OrderTimeInForce.DAY,
		# 		order_type=order.type,	
		# 		legs=new_order_legs,
		# 		price=tasty_price
		# 	)
		# elif order.type == OrderType.STOP:
		# 	replacement_order = NewOrder(
		# 		time_in_force=OrderTimeInForce.DAY,
		# 		order_type=order.type,	
		# 		legs=new_order_legs,
		# 		stop_trigger=tasty_price
		# 	)
		# elif order.type == OrderType.MARKET:
		# 	replacement_order = NewOrder(
		# 		time_in_force=OrderTimeInForce.DAY,
		# 		order_type=order.type,	
		# 		legs=new_order_legs
		# 	)

		try:
			self._replacedOrders.append(tasty_order)  # Merken für das Cancel Event dieser Order
			response: PlacedOrder = account.replace_order(self._session, tasty_order.id, replacement_order)
			order.brokerSpecific['order'] = response
			#self._replacedOrders.append(response) # Auch die neue Order zu den zu ignorierenden Orders hinzufügen
			logger.debug(f'Replacment order {response.id} submitted successfully')
			return True
		
		except TastytradeError as tastyErr:
			logger.error(f'Error adjusting order: {tastyErr}')
			return False
		except ValidationError as valErr:
			logger.error(f'Validation error adjusting order: {valErr}')
			err = valErr.errors()[0]
			logger.error(err)
			return False 

	async def _on_alert_streamer_disconnect(self, streamer: AlertStreamer):
		"""
		Callback method which is called when the Tastytrade alert streamer disconnects.
		Reconnects only the alert streamer without affecting the DXLink streamer.
		"""
		try:
			logger.debug("Tastytrade Alert Streamer disconnected.")
			
			# Check if we're already disconnecting OR if the DXLink streamer is also disconnected
			# (indicating a full disconnect scenario, not just an Alert Streamer issue)
			if self._is_disconnecting:
				logger.debug("Already disconnecting, skipping Alert Streamer reconnect")
				return
			
			if self._streamer is None:
				logger.debug("DXLink streamer is also disconnected (None) - this is a full disconnect, not an Alert Streamer reconnect scenario")
				return
			
			if self._is_disconnecting == False:
				logger.warning("Alert Streamer disconnected unexpectedly. Reconnecting Alert Streamer...")
			
			# CRITICAL: Close the old alert streamer FIRST to stop the async for loop
			# This prevents the old task from receiving events after being cancelled
			old_streamer = self._alert_streamer
			old_streamer_id = id(old_streamer) if old_streamer else None
			self._alert_streamer = None  # Clear reference immediately to prevent race conditions
			logger.debug(f'Set _alert_streamer to None (old streamer was {old_streamer_id})')
			
			if old_streamer:
				try:
					logger.debug(f'Closing old alert streamer {old_streamer_id}')
					await asyncio.wait_for(old_streamer.close(), timeout=2.0)
					logger.debug(f'Old alert streamer {old_streamer_id} closed successfully')
				except asyncio.TimeoutError:
					logger.warning(f'Timeout while closing old alert streamer {old_streamer_id} - continuing anyway')
				except asyncio.CancelledError:
					logger.warning(f'Close operation cancelled for alert streamer {old_streamer_id} - continuing anyway')
				except Exception as exc:
					logger.warning(f'Error closing old alert streamer {old_streamer_id}: {exc} - continuing anyway')
			
			# Now cancel the account listening task (it should exit cleanly now)
			old_task_id = id(self.task_listen_accounts) if self.task_listen_accounts else None
			if self.task_listen_accounts and not self.task_listen_accounts.done():
				logger.debug(f'Cancelling account listening task {old_task_id}')
				self.task_listen_accounts.cancel()
				# Wait for the task using gather to avoid recursion issues
				try:
					await asyncio.wait_for(
						asyncio.gather(self.task_listen_accounts, return_exceptions=True),
						timeout=2.0
					)
					logger.debug(f'Account listening task {old_task_id} cancelled successfully')
				except asyncio.TimeoutError:
					logger.warning(f'Timeout waiting for account listening task {old_task_id} to cancel - continuing anyway')
				except asyncio.CancelledError:
					logger.debug(f'Account listening task {old_task_id} already cancelled')
				except Exception as exc:
					logger.warning(f'Error waiting for account listening task to cancel: {exc} - continuing anyway')
			else:
				logger.debug(f'Account listening task already done or None, no need to cancel')
			
			# Small delay to ensure everything is cleaned up
			logger.debug('Waiting 0.5s for cleanup before reconnecting')
			await asyncio.sleep(0.5)
			
			# Restart the account updates task which will create a new alert streamer
			try:
				# Check if event loop is still running before creating new task
				loop = asyncio.get_event_loop()
				if loop.is_running():
					self.task_listen_accounts = asyncio.create_task(self._request_account_updates())
					new_task_id = id(self.task_listen_accounts)
					logger.info(f"Alert Streamer reconnected successfully (new task {new_task_id})")
				else:
					logger.debug("Event loop not running, skipping Alert Streamer reconnect")
			except RuntimeError as runtime_err:
				logger.warning(f'Cannot reconnect Alert Streamer: {runtime_err}')
			except Exception as exc:
				logger.error(f'Error reconnecting Alert Streamer: {exc}')
		except Exception as e:
			logger.error(f"Critical error in _on_alert_streamer_disconnect: {e}", exc_info=True)

	async def _on_streamer_disconnect(self, streamer: DXLinkStreamer):
		"""
		Callback method which is called when the Tastytrade streamer disconnects.
		
		OTB-236: This callback may be invoked by the tastytrade library during shutdown
		when the event loop is already closing. We need to:
		1. Check if event loop is still running
		2. Avoid calling _disconnect_internal() which would try to close the streamer again (recursion!)
		3. Handle unexpected disconnects without causing infinite loops
		"""
		try:
			logger.debug(f"Tastytrade Streamer disconnect callback triggered (_is_disconnecting={self._is_disconnecting})")
			
			# Check if we're already in the process of disconnecting
			if self._is_disconnecting:
				logger.debug("Already disconnecting, ignoring streamer disconnect callback")
				return
			
			# OTB-236: Check if event loop is still running to avoid RuntimeError during shutdown
			try:
				loop = asyncio.get_running_loop()
				# Note: is_closing() is not available in Python 3.11, only check if loop exists
				logger.debug(f"Event loop is running")
			except RuntimeError as e:
				logger.debug(f"Event loop not running, ignoring streamer disconnect callback (shutdown in progress): {e}")
				return
			
			# Handle unexpected disconnect (not during our controlled shutdown)
			logger.warning("Unexpected DXLink streamer disconnect detected. Initiating full reconnection...")
			logger.debug(f"Streamer: {id(self._streamer) if self._streamer else None}, Alert Streamer: {id(self._alert_streamer) if self._alert_streamer else None}")
			
			# OTB-236: Set flag BEFORE any operations to prevent recursive calls
			self._is_disconnecting = True
			
			# OTB-236: Don't call _disconnect_internal() here as it will try to close the streamer again!
			# Instead, manually clean up and trigger reconnection via the disconnected event
			
			# Disable trading immediately
			await self.set_trading_enabled(False, "Streamer disconnected unexpectedly")
			
			# Cancel streaming tasks - use gather with return_exceptions to avoid recursion
			tasks_to_cancel = [
				task for task in [
					self.task_listen_quotes,
					self.task_listen_greeks,
					self.task_listen_candle,
					self.task_listen_accounts
				] if task and not task.done()
			]
			
			if tasks_to_cancel:
				# Cancel all tasks
				for task in tasks_to_cancel:
					task.cancel()
				
				# Wait for cancellation to complete
				try:
					await asyncio.wait_for(
						asyncio.gather(*tasks_to_cancel, return_exceptions=True),
						timeout=2.0
					)
					logger.debug(f'All {len(tasks_to_cancel)} streaming tasks cancelled')
				except asyncio.TimeoutError:
					logger.debug(f'Timeout waiting for streaming tasks to cancel')
				except Exception as e:
					logger.debug(f'Error waiting for streaming tasks to cancel: {e}')
			
			# Clear task references
			self.task_listen_quotes = None
			self.task_listen_greeks = None
			self.task_listen_candle = None
			self.task_listen_accounts = None
			
			# Clear streamer references (don't call close() - it's already closing!)
			self._streamer = None
			self._alert_streamer = None
			
			# Emit the disconnected event to trigger reconnection logic
			self._emitDisconnectedEvent()
		except Exception as e:
			logger.error(f"Unexpected error in _on_streamer_disconnect: {e}", exc_info=True)
	
	async def _subscribe_data(self):
		"""
		Subscribe to the required data
		"""
		logger.debug("Subscribing to streaming data")
		await self._streamer.subscribe(Quote, self._quote_symbols)
		await self._streamer.subscribe(Greeks, self._greeks_symbols)
		startTime = dt.datetime.now() - timedelta(days=1)
		await self._streamer.subscribe_candle(self._candle_symbols, interval='1m', start_time=startTime)

	async def requestTickerData(self, symbols: List[str]) -> None:
		""" 
		Request ticker data for the given symbols and their options
		"""
		# Check if this is the first call (initial setup) or a re-subscription (e.g., after EOD settlement)
		is_initial_setup = self._streamer is None
		if is_initial_setup:
			ssl_context = ssl.create_default_context(cafile=certifi.where())
			self._streamer = await DXLinkStreamer(self._session, ssl_context=ssl_context, disconnect_fn=self._on_streamer_disconnect)

			self._quote_symbols = []
			self._candle_symbols = []
			self._greeks_symbols = symbols

			# Initial setup: Create new symbolData objects
			for symbol in symbols:
				match symbol:
					case 'SPX':
						symbolData = TastySymbolData()
						symbolData.symbol = symbol
						symbolData.tastySymbol = 'SPX'
						self._quote_symbols.append('SPX')
						self._symbolData[symbol] = symbolData
						self._symbolReverseLookup[symbolData.tastySymbol] = symbol
					case 'VIX':
						symbolData = TastySymbolData()
						symbolData.symbol = symbol
						symbolData.tastySymbol = 'VIX'
						symbolData.trade_options = False  # No options for VIX
						self._candle_symbols.append('VIX')
						self._symbolData[symbol] = symbolData
						self._symbolReverseLookup[symbolData.tastySymbol] = symbol
					case _:
						logger.error(f'Symbol {symbol} currently not supported by Tastytrade Connector!')
						continue

			# Request option chains and expirations for symbols with options trading enable
			current_date = dt.date.today()
			for item in self._symbolData.values():
				symbol_data: TastySymbolData = item
				if symbol_data.trade_options:
					# If options are enabled, request the option chain and expirations
					symbolInformation = symbolInfo.symbol_infos[symbol_data.symbol]
					chains = NestedOptionChain.get(self._session, symbol_data.tastySymbol)
					for chain in chains:
						if chain.root_symbol == symbolInformation.trading_class:
							break
					assert chain != None
					symbol_data.chain = chain

					got_strikes = False
					for expiration in chain.expirations:
						if expiration.expiration_date < current_date:
							continue
						symbol_data.expirations.append(expiration.expiration_date)
						if got_strikes == False:
							# Obtain strikes of the first expiration only
							for strike in expiration.strikes:
								symbol_data.strikes.append(float(strike.strike_price))
							got_strikes = True

					if current_date > symbol_data.expirations[0]:
						logger.warning(f'There are no {symbol_data.symbol} options expiring today!')

			# OTB-253: Symbol data is now ready for trade recovery to use
			# Start subscribe operations in background (don't await - they can take 2+ seconds)
			# This prevents blocking trade recovery which needs symbol data but not live streaming yet
			logger.debug('Symbol data loaded, starting streaming subscriptions in background')
			
			# Start subscribe and streaming tasks in background
			asyncio.create_task(self._start_streaming_async())
		else:
			# Re-subscription after EOD settlement: Keep existing symbolData, just re-subscribe
			logger.debug('Re-subscribing to ticker data after EOD settlement (keeping existing symbolData)')
			
			# Re-subscribe to base symbols (SPX, VIX) - these should still be in the lists
			await self._subscribe_data()
			
			# Mark ticker data as subscribed again
			self._subscribed_ticker_data = True
			logger.debug('Ticker data re-subscription completed')

	async def _start_streaming_async(self):
		"""
		OTB-253: Start streaming subscriptions and tasks asynchronously.
		This method runs in the background to avoid blocking trade recovery.
		"""
		try:
			# Subscribe to data (this can take 2+ seconds due to network calls)
			await self._subscribe_data()
			self._subscribed_ticker_data = True
			
			# OTB-xxx: Start streaming tasks but don't wait for them (they run indefinitely)
			# These tasks will be cancelled during disconnect via _disconnect_internal()
			if self.task_listen_quotes == None:
				self.task_listen_quotes = asyncio.create_task(self._update_quotes())
				self.task_listen_greeks = asyncio.create_task(self._update_greeks())
				self.task_listen_candle = asyncio.create_task(self._update_candle())
				logger.debug('Started streaming tasks for quotes, greeks, and candles')
		except Exception as e:
			logger.error(f'Failed to start streaming: {e}', exc_info=True)

	async def _emit_commission_report_delayed(self, generic_order: GenericOrder, placed_order: PlacedOrder):
		"""
		OTB-267: Emits commission reports for filled Tastytrade orders after a delay.
		
		Tastytrade doesn't provide commission data in the order fill event. We need to query
		the transaction history API after a delay to get regulatory fees and clearing fees.
		
		Args:
			generic_order: The generic order object
			placed_order: The Tastytrade PlacedOrder with order ID
		"""
		try:
			# Wait a bit for fees to be posted to account history (typically 1-2 seconds)
			await asyncio.sleep(3)
			
			# Get account for this order
			for tasty_account in self._tasty_accounts:
				if tasty_account.account_number == placed_order.account_number:
					# Fetch recent transaction history
					from datetime import date, timedelta
					start_date = date.today() - timedelta(days=1)
					transactions = tasty_account.get_history(self._session, start_date=start_date)
					
					# Find transactions matching this order ID
					# Tastytrade Transaction objects have 'id' but PlacedOrder has different 'id'
					# We need to match by symbol, time, and price
					for i, broker_leg in enumerate(placed_order.legs):
						if not broker_leg.fills:
							continue
						
						generic_leg = generic_order.legs[i] if i < len(generic_order.legs) else None
						if not generic_leg:
							continue
						
						# Find matching transaction by fill details
						fill_id = broker_leg.fills[0].fill_id
						total_fees = 0.0
						total_commission = 0.0
						
						# Look for transactions with matching symbol and recent timestamp
						for txn in transactions:
							# Check if transaction matches this leg (by symbol and recent time)
							if (txn.symbol == broker_leg.symbol and 
								txn.instrument_type.value == "Equity Option"):
								# Add up fees
								if hasattr(txn, 'regulatory_fees') and txn.regulatory_fees:
									total_fees += float(txn.regulatory_fees)
								if hasattr(txn, 'clearing_fees') and txn.clearing_fees:
									total_fees += float(txn.clearing_fees)
								# Tastytrade has $0 commission for options
								total_commission = 0.0
						
						# Emit commission report event for this leg
						if total_fees > 0 or total_commission > 0:
							logger.debug(f'Emitting commission report for order {placed_order.id} leg {i}: fee=${total_fees}, commission=${total_commission}')
							self._emitCommissionReportEvent(generic_order, fill_id, total_commission, total_fees)
					
					break
					
		except Exception as e:
			logger.error(f'Failed to emit commission report for order {placed_order.id}: {e}', exc_info=True)

	def _emit_execution_details_for_order(self, generic_order: GenericOrder, placed_order: PlacedOrder):
		"""
		OTB-267: Emits execution details event for filled Tastytrade orders to create transactions.
		This is necessary because Tastytrade API doesn't provide separate execution detail events
		like IBKR does. Without this, trades are created but transactions table remains empty.
		
		The Tastytrade PlacedOrder contains a 'fills' array for each leg with individual fill
		prices and quantities. We process these fills to create execution detail events.
		
		Args:
			generic_order: The generic order object with legs information
			placed_order: The Tastytrade PlacedOrder with fill details per leg
		"""
		from optrabot.broker.order import Execution, OrderAction as GenericOrderAction
		
		# Match Tastytrade legs with our generic order legs
		# PlacedOrder.legs contains the broker-specific leg data with fills
		for i, broker_leg in enumerate(placed_order.legs):
			# Find corresponding generic leg (match by position)
			if i >= len(generic_order.legs):
				logger.warning(f'Broker leg index {i} exceeds generic order legs count {len(generic_order.legs)}')
				continue
			
			generic_leg = generic_order.legs[i]
			
			# Check if leg has fill data
			if not broker_leg.fills or len(broker_leg.fills) == 0:
				logger.warning(f'No fill data for leg {i} of order {placed_order.id}')
				# Fallback: Use order price divided by number of legs
				fill_price = abs(float(placed_order.price)) / len(placed_order.legs) if placed_order.price else 0.0
				fill_quantity = float(broker_leg.quantity) if broker_leg.quantity else 0.0
				fill_id = f"{placed_order.id}-{i}"
				fill_timestamp = dt.datetime.now()
			else:
				# Use actual fill data from Tastytrade API
				# Sum up all fills for this leg (can be partial fills)
				total_fill_quantity = sum(float(fill.quantity) for fill in broker_leg.fills)
				# Calculate weighted average fill price
				total_fill_value = sum(float(fill.quantity) * float(fill.fill_price) for fill in broker_leg.fills)
				fill_price = abs(total_fill_value / total_fill_quantity) if total_fill_quantity > 0 else 0.0
				fill_quantity = total_fill_quantity
				# Use first fill ID and timestamp
				fill_id = broker_leg.fills[0].fill_id
				fill_timestamp = broker_leg.fills[0].filled_at
			
			# Determine action: BUY_TO_OPEN/CLOSE -> BUY, SELL_TO_OPEN/CLOSE -> SELL
			if generic_leg.action in [GenericOrderAction.BUY_TO_OPEN, GenericOrderAction.BUY_TO_CLOSE, GenericOrderAction.BUY]:
				action = GenericOrderAction.BUY
			else:
				action = GenericOrderAction.SELL
			
			# Create execution object with actual fill details
			execution = Execution(
				id=fill_id,
				action=action,
				sec_type=generic_leg.right.value,  # 'Call' or 'Put'
				strike=generic_leg.strike,
				amount=int(fill_quantity),
				price=fill_price,  # Actual fill price from Tastytrade API
				expiration=generic_leg.expiration,
				timestamp=fill_timestamp
			)
			
			logger.debug(f'Emitting execution details for Tastytrade order {placed_order.id}: {generic_leg.symbol} {generic_leg.right} {generic_leg.strike} qty={execution.amount} @ ${execution.price}')
			self._emitOrderExecutionDetailsEvent(generic_order, execution)

	async def _update_accounts(self):
		"""
		Task for listening to account updates
		"""
		logger.debug(f'_update_accounts started, alert_streamer={id(self._alert_streamer) if self._alert_streamer else None}')
		try:
			# Check if streamer is still valid before entering the loop
			if self._alert_streamer is None:
				logger.debug('Alert streamer is None, cannot start account updates loop')
				return
			
			streamer_id = id(self._alert_streamer)
			logger.debug(f'Starting async for loop with streamer {streamer_id}')
			async for order in self._alert_streamer.listen(PlacedOrder):
				logger.debug(f'Received order event in loop (streamer {streamer_id}): {order.id} status {order.status}')
				additional_info = ''
				if order.status == OrderStatus.REJECTED:
					additional_info = f'Reason: {order.reject_reason}'
				logger.debug(f'Update on order {order.id} status {order.status} {additional_info}')			
				ignore_order_event = False
				# Cancel Events von Preisanpassungen ignorieren, da sie kein echtes Cancel sind
				for replaced_order in self._replacedOrders:
					if replaced_order.id == order.id and order.status == OrderStatus.CANCELLED:
						#self._replacedOrders.remove(replaced_order)
						ignore_order_event = True
						logger.debug('Ignoring cancel event for replaced order')
						continue
					if replaced_order.id == order.id and (order.status == OrderStatus.ROUTED or order.status == OrderStatus.LIVE):
						if order.status == OrderStatus.LIVE:
							self._replacedOrders.remove(replaced_order)
						ignore_order_event = True
						logger.debug('Ignoring placement of new replacement order')
						continue
		
				if ignore_order_event == False:
					relevantOrder: GenericOrder = None
					for managedOrder in self._orders:
						broker_specific_order: PlacedOrder = managedOrder.brokerSpecific['order']
						if broker_specific_order.id == order.id:
							relevantOrder = managedOrder
							break
				
					if relevantOrder == None:
						logger.debug(f'No managed order matched the status event')
					else:
						relevantOrder.brokerSpecific['order'] = order
						filledAmount = int(order.size)
						relevantOrder.averageFillPrice = abs(float(order.price)) if order.price != None else 0
						order_status =	self._genericOrderStatus(order.status)
						if order_status != None:
							self._emitOrderStatusEvent(relevantOrder, order_status, filledAmount)
							# Emit execution details when order is filled
							if order.status == OrderStatus.FILLED:
								logger.debug(f'Order {order.id} is filled, emitting execution details')
								self._emit_execution_details_for_order(relevantOrder, order)
						else:
							logger.debug(f'Order status {order.status} not mapped to generic order status')
		except asyncio.CancelledError:
			logger.debug(f'Account updates task cancelled, stopping order listener (streamer {id(self._alert_streamer) if self._alert_streamer else None})')
			raise  # Re-raise to allow proper task cancellation
		except Exception as exc:
			logger.error(f'Unexpected error in _update_accounts: {exc}', exc_info=True)
		finally:
			logger.debug(f'_update_accounts exiting (streamer {id(self._alert_streamer) if self._alert_streamer else None})')

	async def _update_quotes(self):
		async for e in self._streamer.listen(Quote):
			logger.trace(f'Received Quote: {e.event_symbol} bid price: {e.bid_price} ask price: {e.ask_price}')
			# Preisdaten speichern if the broker is subscribed to ticker data
			if self._subscribed_ticker_data is None:
				continue

			if not e.event_symbol.startswith('.'):
				# Symbol ist ein Basiswert
				try:
					genericSymbol = self._symbolReverseLookup[e.event_symbol]
					symbolData: TastySymbolData  = self._symbolData[genericSymbol]
					midPrice = float((e.bid_price + e.ask_price) / 2)
					atmStrike = OptionHelper.roundToStrikePrice(midPrice)
					symbolData.lastPrice = midPrice
					if symbolData.lastAtmStrike != atmStrike:  # Check for missing Option Data only if ATM Strike has changed
						symbolData.lastAtmStrike = atmStrike
						asyncio.create_task(self._requestMissingOptionData(symbolData, atmStrike))
					
				except KeyError as keyErr:
					logger.error(f'No generic symbol found for tastytrade symbol {e.event_symbol}')
			else:
				# Symbol ist eine Option
				try:
					symbol, optionType, expiration, strike = self._getOptionInfos(e.event_symbol)
					symbol_information = symbolInfo.symbol_infos[symbol]
					symbolData = self._symbolData[symbol]
					optionStrikeData = symbolData.optionPriceData[expiration]
					optionStrikePriceData = optionStrikeData.strikeData[strike]
					if optionType == OptionType.CALL:
						optionStrikePriceData.callBid = float(e.bid_price)
						optionStrikePriceData.callAsk = float(e.ask_price)
					else:
						optionStrikePriceData.putBid = float(e.bid_price)
						optionStrikePriceData.putAsk = float(e.ask_price)
					optionStrikePriceData.lastUpdated = dt.datetime.now(symbol_information.timezone)
					self._last_option_price_update_time = optionStrikePriceData.lastUpdated
				except Exception as exc:
					logger.error(f'Error getting option infos: {exc}')
	
	async def _update_greeks(self):
		async for e in self._streamer.listen(Greeks):
			logger.trace(f'Received Greeks: {e.event_symbol} delta: {e.delta}')
			if e.event_symbol.startswith('.'):
				# Symbol ist eine Option
				try:
					symbol, optionType, expiration, strike = self._getOptionInfos(e.event_symbol)
					symbol_information = symbolInfo.symbol_infos[symbol]
					symbolData = self._symbolData[symbol]
					optionStrikeData = symbolData.optionPriceData[expiration]
					optionStrikePriceData = optionStrikeData.strikeData[strike]
					if optionType == OptionType.CALL:
						optionStrikePriceData.callDelta = float(e.delta)
					else:
						optionStrikePriceData.putDelta = float(e.delta)

				except Exception as exc:
					logger.error(f'Error getting option infos: {exc}')

	async def _update_candle(self):
		async for e in self._streamer.listen(Candle):
			logger.trace(f'Received Candle: {e.event_symbol} close: {e.close}')
			if not e.event_symbol.startswith('.'):
				# Symbol ist ein Basiswert
				try:
					symbol = e.event_symbol.split('{')[0]
					genericSymbol = self._symbolReverseLookup[symbol]
					symbolData: TastySymbolData  = self._symbolData[genericSymbol]
					symbolData.lastPrice = float(e.close)
				except KeyError as keyErr:
					logger.error(f'No generic symbol found for tastytrade symbol {e.event_symbol}')

	async def eod_settlement_tasks(self):
		"""
		Perform End of Day settlement tasks
		"""
		await super().eod_settlement_tasks()

	def get_option_strike_data(self, symbol: str, expiration: dt.date) -> OptionStrikeData:
		""" 
		Returns the option strike data for the given symbol and expiration. It is including
		prices and greeks.
		"""
		symbolData = self._symbolData[symbol]
		try:
			return symbolData.optionPriceData[expiration]
		except KeyError:
			raise ValueError(f'No option strike data for symbol {symbol} and expiration {expiration} found!')

	def get_option_strike_price_data(self, symbol: str, expiration: date, strike: float) -> OptionStrikePriceData:
		""" 
		Returns the option strike price data for the given symbol, expiration date, strike price and right.
		Returns None if no price data is available for the expiration date or strike.
		"""
		symbolData = self._symbolData[symbol]
		try:
			optionStrikeData = symbolData.optionPriceData[expiration]
		except KeyError:
			# No price data available for this expiration date (e.g., after trade recovery)
			return None
		
		if strike in optionStrikeData.strikeData.keys():
			return optionStrikeData.strikeData[strike]
		else:
			return None

	def get_strike_by_delta(self, symbol: str, right: str, delta: int) -> float:
		"""
		Returns the strike price based on the given delta based on the buffered option price data
		"""
		symbolData = self._symbolData[symbol]
		current_date = dt.date.today()
		option_price_data: OptionStrikeData= symbolData.optionPriceData[current_date]
		previous_delta = 0
		previous_strike = 0
		reverse = True if right == OptionRight.PUT else False
		sorted_strikes = dict(sorted(option_price_data.strikeData.items(), reverse=reverse))
		for strike, price_data in sorted_strikes.items():
			if right == OptionRight.PUT:
				if price_data.putDelta == None:
					continue
				adjusted_delta = price_data.putDelta * -100
			else:
				if price_data.callDelta == None:
					continue
				adjusted_delta = price_data.callDelta * 100
			if adjusted_delta <= delta:
				if OptionHelper.closest_number(delta, previous_delta, adjusted_delta) == adjusted_delta:
					return strike
				else:
					return previous_strike
			previous_delta = adjusted_delta
			previous_strike = strike

		raise ValueError(f'No strike price found for delta {delta} in symbol {symbol}!')
	
	def get_strike_by_price(self, symbol: str, right: str, price: float) -> float:
		""" 
		Returns the strike price based on the given premium price based on the buffered option price data
		"""
		# TODO: Implement in base class, because TWS Connector got the same code
		symbolData = self._symbolData[symbol]
		current_date = dt.date.today()
		option_price_data: OptionStrikeData= symbolData.optionPriceData[current_date]
		previous_price = 0
		previous_strike = 0
		reverse = True if right == OptionRight.PUT else False
		sorted_strikes = dict(sorted(option_price_data.strikeData.items(), reverse=reverse))
		for strike, price_data in sorted_strikes.items():
			if right == OptionRight.PUT:
				current_strike_price = price_data.getPutMidPrice()
				if current_strike_price == None:
					continue
			elif right == OptionRight.CALL:
				current_strike_price = price_data.getCallMidPrice()
				if current_strike_price == None:
					continue
			if current_strike_price > 0 and current_strike_price <= price:
				if OptionHelper.closest_number(price, previous_price, current_strike_price) == current_strike_price:
					return strike
				else:
					return previous_strike
			previous_price = current_strike_price
			previous_strike = strike
		raise ValueError(f'No strike price found for price {price} in symbol {symbol}!')

	def getFillPrice(self, order: GenericOrder) -> float:
		""" 
		Returns the fill price of the given order if it is filled
		"""
		try:
			tastyOrder: PlacedOrder = order.brokerSpecific['order']
			if tastyOrder.status == OrderStatus.FILLED:
				return abs(float(tastyOrder.price))
			else:
				return 0
		except KeyError as keyErr:
			logger.error(f'No fill price available for order {order}')
	
	def getLastPrice(self, symbol: str) -> float:
		""" 
		Returns the last price of the given symbol
		"""
		try:
			symbolData = self._symbolData[symbol]
			return symbolData.lastPrice
		except KeyError as keyErr:
			logger.error(f'No last price available for symbol {symbol}')
			return 0

	def oco_as_complex_order(self) -> bool:
		"""
		With Tastytrade, the OCO orders have to be placed as one complex order
		"""
		return True
	
	def uses_oco_orders(self) -> bool:
		""" 
		The TWS Connector uses OCO orders for take profit and stop loss orders
		"""
		return True
	
	async def unsubscribe_ticker_data(self):
		"""
		Unsubscribe from ticker data - including base symbols and all options
		"""
		await super().unsubscribe_ticker_data()
		
		# First, unsubscribe from all option symbols
		for symbolData in self._symbolData.values():
			try:
				for day_option_price_data in symbolData.optionPriceData.values():
					streamer_symbols = []
					for value in day_option_price_data.strikeData.values():
						option_price_data: OptionStrikePriceData = value
						
						try:
							call_streamer_symbol = option_price_data.brokerSpecific['call_streamer_symbol']
							streamer_symbols.append(call_streamer_symbol)
							self._quote_symbols.remove(call_streamer_symbol)
						except KeyError as keyErr:
							pass

						try:
							put_streamer_symbol = option_price_data.brokerSpecific['put_streamer_symbol']
							streamer_symbols.append(put_streamer_symbol)
							self._quote_symbols.remove(put_streamer_symbol)
						except KeyError as keyErr:
							pass
				
					await self._streamer.unsubscribe(Quote, streamer_symbols)
					await self._streamer.unsubscribe(Greeks, streamer_symbols)
				symbolData.optionPriceData.clear()
				symbolData.lastAtmStrike = 0
				symbolData.lastPrice = 0
			except KeyError as keyErr:
				pass
		
		# Now unsubscribe from base symbols (SPX, VIX, etc.)
		# This ensures a clean state for re-subscription at new day start
		if self._quote_symbols:
			logger.debug(f'Unsubscribing from {len(self._quote_symbols)} base quote symbols')
			await self._streamer.unsubscribe(Quote, self._quote_symbols)
		if self._greeks_symbols:
			logger.debug(f'Unsubscribing from {len(self._greeks_symbols)} base greeks symbols')
			await self._streamer.unsubscribe(Greeks, self._greeks_symbols)
		if self._candle_symbols:
			logger.debug(f'Unsubscribing from {len(self._candle_symbols)} base candle symbols')
			await self._streamer.unsubscribe(Candle, self._candle_symbols)

	async def _requestMissingOptionData(self, symbolData: TastySymbolData, atmStrike: float):
		"""
		Request option data for the given symbol and expiration date
		"""
		relevant_expirations = self._determine_relevant_expirations(symbolData)
		symbolInformation = symbolInfo.symbol_infos[symbolData.symbol]
		strikes_of_interest = self._determine_strikes_of_interest(symbolData, atmStrike)	
	
		for relevant_expiration in relevant_expirations:
			# Optain saved options data for the expiration date
			try:
				#expiration_date_str = relevant_expiration.strftime('%Y%m%d')
				optionStrikeData = symbolData.optionPriceData[relevant_expiration]
			except KeyError as keyErr:
				# Wenn noch keine Optionsdaten für das Verfallsdatum vorhanden sind, dann bei Tasty anfragen ob es Optionsdaten gibt

				optionStrikeData = OptionStrikeData()
				symbolData.optionPriceData[relevant_expiration] = optionStrikeData

			for chain_at_expiration in symbolData.chain.expirations:
				if chain_at_expiration.expiration_date >= relevant_expiration:
					break
			if chain_at_expiration == None or chain_at_expiration.expiration_date != relevant_expiration:
				logger.error(f'No options available for symbol {symbolData.tastySymbol} and expiration date {relevant_expiration}')
				continue
			
			# Convert the available strikes to a list of decimal numbers
			# if not done yet
			try:
				available_strikes = chain_at_expiration._optrabot_strikes
			except AttributeError as attrErr:
				chain_at_expiration._optrabot_strikes = []
				available_strikes = chain_at_expiration._optrabot_strikes
				for strike in chain_at_expiration.strikes:
					available_strikes.append(strike.strike_price)

			options_to_be_requested = []
			for strike_price in strikes_of_interest:	
				try:
					optionStrikeData.strikeData[strike_price]
				except KeyError as keyErr:
					option_strike_price_data = OptionStrikePriceData()
					optionStrikeData.strikeData[strike_price] = option_strike_price_data
					options_to_be_requested.append(strike_price)

			if len(options_to_be_requested) > 0:
				streamer_symbols = []
				for item in chain_at_expiration.strikes:
					strike: Strike = item
					if strike.strike_price in options_to_be_requested:
						option_strike_data = optionStrikeData.strikeData[strike.strike_price]
						option_strike_data.brokerSpecific['call_option'] = strike.call
						option_strike_data.brokerSpecific['call_streamer_symbol'] = strike.call_streamer_symbol
						option_strike_data.brokerSpecific['put_option'] =  strike.put
						option_strike_data.brokerSpecific['put_streamer_symbol']  = strike.put_streamer_symbol
						streamer_symbols.append(strike.call_streamer_symbol)
						streamer_symbols.append(strike.put_streamer_symbol)
						self._quote_symbols.append(strike.call_streamer_symbol)
						self._quote_symbols.append(strike.put_streamer_symbol)
				await self._streamer.subscribe(Quote, streamer_symbols)
				await self._streamer.subscribe(Greeks, streamer_symbols)
		
	def _getOptionInfos(self, tastySymbol: str) -> tuple:
		"""
		Extracts the generic symbol and expiration date, strike and option side from the tastytrade option symbol.
		If the option symbol information cannot be parsed as expected, a ValueError exception is raised.
		"""
		error = False
		pattern = r'^.(?P<optionsymbol>[A-Z]+)(?P<expiration>[0-9]+)(?P<type>[CP])(?P<strike>[0-9]+)'
		compiledPattern = re.compile(pattern)
		match = compiledPattern.match(tastySymbol)
		try:
			if match:
				optionSymbol = match.group('optionsymbol')
				for symbol, symbol_info in symbolInfo.symbol_infos.items():
					if symbol_info.symbol + symbol_info.option_symbol_suffix == optionSymbol:
						genericSymbol = symbol_info.symbol
						break
				expirationDate = dt.datetime.strptime(match.group('expiration'), '%y%m%d').date()
				strike = float(match.group('strike'))
				optionType = OptionType.CALL if match.group('type') == 'C' else OptionType.PUT
		except IndexError as indexErr:
			logger.error(f'Invalid option symbol {tastySymbol}')
			error = True
		except ValueError as valueErr:
			logger.error(f'Invalid option symbol {tastySymbol}')
			error = True
		if genericSymbol == None or error == True:
			raise ValueError(f'Invalid option symbol {tastySymbol}')
		return genericSymbol, optionType, expirationDate, strike
	
	def _mappedOrderAction(self, orderAction: GenericOrderAction, legAction: GenericOrderAction) -> OrderAction:
		"""
		Maps the general order action to the Tasty specific order action
		For Closing orders, the leg actions need to be inverted
		eg. BUY -> SELL_TO_CLOSE
		eg. SELL -> BUY_TO_CLOSE
		"""
		match orderAction:
			case GenericOrderAction.BUY_TO_OPEN:
				if legAction == GenericOrderAction.BUY:
					return OrderAction.BUY_TO_OPEN
				if legAction == GenericOrderAction.SELL:
					return OrderAction.SELL_TO_OPEN
				else:
					raise ValueError(f'Unknown leg action: {legAction}')
			case GenericOrderAction.SELL_TO_OPEN:
				if legAction == GenericOrderAction.SELL:
					return OrderAction.SELL_TO_OPEN
				elif legAction == GenericOrderAction.BUY:
					return OrderAction.BUY_TO_OPEN
				else:
					raise ValueError(f'Unknown leg action: {legAction}')
			case GenericOrderAction.BUY_TO_CLOSE:
				if legAction == GenericOrderAction.BUY:
					return OrderAction.BUY_TO_CLOSE
				if legAction == GenericOrderAction.SELL:
					return OrderAction.SELL_TO_CLOSE
				else:
					raise ValueError(f'Unknown leg action: {legAction}')
			case GenericOrderAction.SELL_TO_CLOSE:
				if legAction == GenericOrderAction.SELL:
					return OrderAction.SELL_TO_CLOSE
				elif legAction == GenericOrderAction.BUY:
					return OrderAction.BUY_TO_CLOSE
				else:
					raise ValueError(f'Unknown leg action for SELL_TO_CLOSE: {legAction}')
			case _:
				raise ValueError(f'Unknown order action: {orderAction}')
			
	async def _request_account_updates(self):
		"""
		Request Account Updates - runs as a background task
		"""
		task_id = id(asyncio.current_task())
		logger.debug(f'_request_account_updates started (task {task_id})')
		try:
			self._alert_streamer = await AlertStreamer(self._session, disconnect_fn=self._on_alert_streamer_disconnect)
			streamer_id = id(self._alert_streamer)
			logger.debug(f'Created alert streamer {streamer_id} (task {task_id})')
			await self._alert_streamer.subscribe_accounts(self._tasty_accounts)
			
			# Start the update loop directly without creating nested tasks
			logger.debug(f'Starting _update_accounts (task {task_id}, streamer {streamer_id})')
			await self._update_accounts()
		except asyncio.CancelledError:
			logger.debug(f'Cancelled listening to account updates (task {task_id})')
			raise  # Re-raise to allow proper task cancellation
		except Exception as exc:
			logger.error(f'Error in account updates (task {task_id}): {exc}')
		finally:
			logger.debug(f'_request_account_updates exiting (task {task_id})')

	def _genericOrderStatus(self, status: OrderStatus) -> GenericOrderStatus:
		"""
		Maps the Tastytrade order status to the generic order status
		"""
		match status:
			case OrderStatus.RECEIVED:
				return GenericOrderStatus.OPEN
			case OrderStatus.LIVE:
				return GenericOrderStatus.OPEN
			case OrderStatus.CONTINGENT:
				return GenericOrderStatus.OPEN
			case OrderStatus.CANCELLED:
				return GenericOrderStatus.CANCELLED
			case OrderStatus.FILLED:
				return GenericOrderStatus.FILLED
			case OrderStatus.REJECTED:
				return GenericOrderStatus.CANCELLED
			case _:
				return None


"""Add trade_group_id, openDate and closeDate to trades table

Revision ID: a1b2c3d4e5f6
Revises: cc3c6f4d83dc
Create Date: 2025-10-23 17:00:00.000000

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = 'a1b2c3d4e5f6'
down_revision: Union[str, None] = 'cc3c6f4d83dc'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Add trade_group_id, openDate and closeDate columns to trades table"""
    # Add trade_group_id column (nullable, for grouping related trades like rollovers)
    op.add_column('trades', sa.Column('trade_group_id', sa.String(), nullable=True))
    
    # Add openDate column (nullable, timestamp when trade was opened)
    op.add_column('trades', sa.Column('openDate', sa.TIMESTAMP(), nullable=True))
    
    # Add closeDate column (nullable, timestamp when trade was closed)
    op.add_column('trades', sa.Column('closeDate', sa.TIMESTAMP(), nullable=True))
    
    # Migrate existing trade data: Set openDate and closeDate based on transactions
    connection = op.get_bind()
    
    # Update openDate: Set to timestamp of first transaction (MIN)
    connection.execute(sa.text("""
        UPDATE trades
        SET openDate = (
            SELECT MIN(timestamp)
            FROM transactions
            WHERE transactions.tradeid = trades.id
        )
        WHERE EXISTS (
            SELECT 1 FROM transactions WHERE transactions.tradeid = trades.id
        )
    """))
    
    # Update closeDate: Set to timestamp of last transaction (MAX), only for CLOSED trades
    connection.execute(sa.text("""
        UPDATE trades
        SET closeDate = (
            SELECT MAX(timestamp)
            FROM transactions
            WHERE transactions.tradeid = trades.id
        )
        WHERE status = 'CLOSED'
        AND EXISTS (
            SELECT 1 FROM transactions WHERE transactions.tradeid = trades.id
        )
    """))


def downgrade() -> None:
    """Remove trade_group_id, openDate and closeDate columns from trades table"""
    op.drop_column('trades', 'closeDate')
    op.drop_column('trades', 'openDate')
    op.drop_column('trades', 'trade_group_id')



import asyncio
import datetime as dt
import re
from typing import Dict, List
from optrabot.exceptions.orderexceptions import PlaceOrderException, PrepareOrderException
from optrabot.broker.optionpricedata import OptionStrikeData, OptionStrikePriceData
from optrabot.optionhelper import OptionHelper
from optrabot.broker.brokerconnector import BrokerConnector, SymbolData
from loguru import logger
from ib_async import *
import optrabot.config as optrabotcfg
from optrabot.models import Account
from optrabot.broker.order import Leg, OptionRight, Order as GenericOrder, OrderAction, OrderType, OrderStatus as GenericOrderStatus, Execution, PriceEffect
from optrabot.tradetemplate.templatefactory import Template
import optrabot.symbolinfo as symbolInfo
from optrabot.managedtrade import ManagedTrade

class IBSymbolData(SymbolData):
	def __init__(self) -> None:
		super().__init__()
		self.contract: Contract = None
		self.option_trading_class: str = None
		self.ticker: Ticker = None

class ForeignOrder(GenericOrder):
	"""
	Represents an order that has been placed by another user in the TWS.
	It is just used for managing conflicting orders.
	"""
	def __init__(self) -> None:
		super().__init__()

class IBTWSTConnector(BrokerConnector):
	def __init__(self) -> None:
		super().__init__()
		self._host = ''
		self._port = None
		self._clientid = None
		self._initialize()
		self.id = 'IBTWS'
		self.broker = 'IBKR'
		self._lock = asyncio.Lock()
		self._ib = IB()
		self._ib.errorEvent += self.onErrorEvent
		self._ib.disconnectedEvent += self.onDisconnected
		self._ib.execDetailsEvent += self.onExecDetailsEvent
		self._ib.orderStatusEvent += self.onOrderStatusEvent
		self._ib.commissionReportEvent += self.onCommissionReportEvent
		self._ib.pendingTickersEvent += self.onPendingTickers
		self._compiled_option_pattern = re.compile(pattern = r'^(?P<optionsymbol>[A-Z]+)\ +(?P<expiration>[0-9]+)(?P<type>[CP])(?P<strike>[0-9]+)')
		self._symbolReverseLookup: Dict[str, str] = {}		# maps IB symbol to generic symbol
		self._orders: List[GenericOrder] = []
		self._duplicate_order_ids: List[int] = []			# Liste mit Order-IDs, die als dopplete Order IDs von TWS gemeldet wurden
		self._conflictingOrderMap: Dict[GenericOrder,GenericOrder] = {}   # Maps the conflicting order to the causing order
		self._rule_details:  Dict[str, List[PriceIncrement]] = {}   # Map of Rule ID to Price Increment rules
		self._lock_rule = asyncio.Lock()
		self._strikes_wo_option: List[str] = []		# List of strikes without option data

	def ___del__(self):
		"""
		Disconnect from TWS when object is deleted
		"""
		if self._ib.isConnected():
			self._ib.disconnect()

		# Unregister event handlers
		self._ib.errorEvent -= self.onErrorEvent
		self._ib.disconnectedEvent -= self.onDisconnected
		self._ib.execDetailsEvent -= self.onExecDetailsEvent
		self._ib.orderStatusEvent -= self.onOrderStatusEvent
		self._ib.commissionReportEvent -= self.onCommissionReportEvent
		self._ib.pendingTickersEvent -= self.onPendingTickers

	def _initialize(self):
		"""
		Initialize the TWS connector from the configuration
		"""
		config :optrabotcfg.Config = optrabotcfg.appConfig
		if config is None:
			return
		try:
			config.get('tws')
		except KeyError as keyErr:
			logger.debug('No TWS connection configured')
			return

		self._host = config.get('tws.host')
		if self._host == '':
			self._host = 'localhost'
		
		try:
			self._port = int(config.get('tws.port'))
		except KeyError as keyErr:
			self._port = 7496
		try:
			self._clientid = int(config.get('tws.clientid'))
		except KeyError as keyErr:
			self._clientid = 21
		self._initialized = True

	async def cancel_order(self, order: GenericOrder):
		""" 
		Cancels the given order
		"""
		await super().cancel_order(order)
		ibTrade: Trade = order.brokerSpecific['trade']
		if ibTrade.orderStatus.status != OrderStatus.Cancelled:
			self._ib.cancelOrder(ibTrade.order)

	async def connect(self):
		await super().connect()
		
		asyncio.create_task(self._connect_tws_task())

	async def disconnect(self):
		await super().disconnect()
		try:
			self._ib.disconnect()
		except Exception as excp:
			pass

	async def eod_settlement_tasks(self):
		"""
		Perform End of Day settlement tasks.
		- Clear list of unavailable options
		"""
		self._strikes_wo_option.clear()

	def getAccounts(self) -> List[Account]:
		"""
		Returns the accounts managed by the TWS connection
		"""
		if len(self._managedAccounts) == 0 and self.isConnected():
			for managedAccount in self._ib.managedAccounts():
				account = Account(id = managedAccount, name = managedAccount, broker = self.broker, pdt = False)
				self._managedAccounts.append(account)
		return self._managedAccounts

	def isConnected(self) -> bool:
		return self._ib.isConnected()
	
	async def prepareOrder(self, order: GenericOrder, need_valid_price_data: bool = True) -> None:
		"""
		Prepares the given order for execution.
		- Retrieve current market data for order legs

		Args:
			order: The order to prepare
			need_valid_price_data: If True, validates price data availability and freshness.
								   If False, creates contracts without price validation (e.g., for trade recovery)

		Raises:
			PrepareOrderException: If the order preparation fails with a specific reason.
		"""
		logger.debug("Wait for lock Prepare Order")
		async with self._lock:
			logger.debug("Got Prepare Order lock")
			symbolInformation = symbolInfo.symbol_infos[order.symbol]
			symbolData = self._symbolData[order.symbol]
			comboLegs: list[ComboLeg] = []
			comboContracts: list[Contract] = []
			for leg in order.legs:
				leg_contract = None
				
				if need_valid_price_data:
					# Normal order flow: use price data with validation
					try:
						option_price_data = symbolData.optionPriceData[leg.expiration]
					except KeyError:
						raise PrepareOrderException(f'No price data for expiration {leg.expiration} available', order)

					try:
						option_strike_price_data = option_price_data.strikeData[leg.strike]
					except KeyError as keyErr:
						raise PrepareOrderException(f'No price data for strike {leg.strike} available', order)
					
					leg_contract = option_strike_price_data.brokerSpecific['call_contract'] if leg.right == OptionRight.CALL else option_strike_price_data.brokerSpecific['put_contract']
					
					if not OptionHelper.checkContractIsQualified(leg_contract):
						raise PrepareOrderException(f'Contract {leg_contract} is not qualified', order)
					
					# Preisdaten für das Order Leg übernehmen
					if not option_strike_price_data.is_outdated():
						leg.askPrice = option_strike_price_data.putAsk if leg.right == OptionRight.PUT else option_strike_price_data.callAsk
						leg.bidPrice = option_strike_price_data.putBid if leg.right == OptionRight.PUT else option_strike_price_data.callBid
					else:
						raise PrepareOrderException(f'Price data for strike {leg.strike} is outdated', order)
				else:
					# Trade recovery flow: create contract without price data validation
					# This is used when recovering trades after restart before market data is available
					expiration_str = leg.expiration.strftime('%Y%m%d')
					leg_contract = Option(
						symbolData.contract.symbol,
						expiration_str,
						leg.strike,
						self._mappedRight(leg.right),
						symbolInformation.exchange,
						tradingClass=symbolData.option_trading_class
					)
					
					# Qualify the contract to get the conId needed for ComboLeg
					try:
						qualified_contracts = await asyncio.wait_for(
							self._ib.qualifyContractsAsync(leg_contract),
							timeout=10
						)
						if qualified_contracts and len(qualified_contracts) > 0:
							leg_contract = qualified_contracts[0]
						else:
							raise PrepareOrderException(f'Failed to qualify contract for strike {leg.strike}', order)
					except asyncio.TimeoutError:
						raise PrepareOrderException(f'Timeout qualifying contract for strike {leg.strike}', order)
					
				comboContracts.append(leg_contract)
				leg.brokerSpecific['contract'] = leg_contract
				comboLeg = ComboLeg(conId=leg_contract.conId, ratio=1, action=self._mapped_leg_action(order.action,leg.action), exchange=symbolInformation.exchange)
				comboLegs.append(comboLeg)
			order.brokerSpecific['comboLegs'] = comboLegs
			
			# Only determine price effect when we have valid price data
			# During recovery, price_effect is already set from transactions
			if need_valid_price_data:
				order.determine_price_effect()

	async def placeOrder(self, managed_trade: ManagedTrade, order: GenericOrder, parent_order: GenericOrder = None) -> None:
		""" 
		Places the given order for a managed account via the broker connection.
		
		Raises:
			PlaceOrderException: If the order placement fails with a specific reason.
		"""
		logger.debug('Wait for Place order lock')
		async with self._lock:
			logger.debug('Got Place order lock')
			symbolInformation = symbolInfo.symbol_infos[order.symbol]
			symbolData = self._symbolData[order.symbol]		
	
			comboLegs = order.brokerSpecific['comboLegs']
			if not comboLegs:
				logger.error("Internal Error: Broker specific attribute comboContracts not filled!")
				return
			
			logger.debug('Checking for open Trades if Combo Legs are colliding')
			#contractsWithOpenOrders = await self._getAllOpenOrderContracts(template.account, symbolData.contract.symbol)

			conflictingOrders = await self._findConflictingOrders(order, managed_trade.template.account)
			if len(conflictingOrders) > 0:
				cancelledOCAs = []
				logger.debug(f'Found {len(conflictingOrders)} conflicting orders for symbol {order.symbol}...adding them to the conflicting order map')
				for conflictingIBTrade in conflictingOrders:
					conflictingOrder = self._findOrderByIBTrade(conflictingIBTrade)
					if conflictingOrder == None:
						raise PlaceOrderException(f'Foreign order from another user at the strike prices', order=order)

					self._conflictingOrderMap[conflictingOrder] = order # Conflicting Order mit der verursachenden Order verknüpfen und merken
					
					# Cancel the conflicting order
					if conflictingIBTrade.order.ocaGroup and conflictingIBTrade.order.ocaGroup in cancelledOCAs:
						# Order braucht nicht storniert werden, da bereits eine andere aus der gleichen OCA Gruppe storniert wurde
						pass
					else:
						self._ib.cancelOrder(conflictingIBTrade.order)
						if conflictingIBTrade.order.ocaGroup:
							cancelledOCAs.append(conflictingIBTrade.order.ocaGroup)

				# Urspüngliche Broker Order - Broker Specific [trade] ; Aktuelle Broker Order ; [Conflicting Trades]

				# Ich stormiere die Conflicting Trades
				# - Beim OrderStatusEvent darf die Stonierung für diese Orders nicht weitergreicht werden, weil sie in der Conflicting Trade Map
				#   enthalten sind

				# Es wird die aktuelle Order platziert
				# - Beim OrderStatusEvent "Filled" muss zusätzlich geprüft werden ob es in der Conflicting Trade Map einen Eintrag gibt
				#   wenn ja, dann müssen die referenzierten Conflicting Trades wieder in den Markt gelegt werden

			# Die Orders der behindernden Trades müssen hier storniert werden und deren Daten zwischengespeichert werden
			# Auch der Bezug zu dieser Order, denn wenn diese ausgeführt worden ist müssen die stornierten Orders wieder neu in den Markt gelegt werden.
			#for trade in conflictingTrades:
			#	trade.order.transmit = False
			#	trade.orderStatus.status = OrderStatus.Inactive
			#	self._ib.placeOrder(trade.contract, trade.order)

			#if len(conflictingTrades) > 0:
			#	logger.error(f'Order with opposite action already open for contract. Order not placed!')
			#	return False

			comboContract = None
			if len(order.legs) > 1:
				comboContract = Contract(symbol=symbolData.contract.symbol, secType='BAG', exchange=symbolInformation.exchange, currency=symbolData.contract.currency, comboLegs=comboLegs)
			else:
				comboContract = order.legs[0].brokerSpecific['contract']

			ib_order_price = self._determine_ib_order_price(order)

			if order.type == OrderType.LIMIT:
				#order.price -= 0.50
				ibOrder = LimitOrder(self._mappedOrderAction(order.action), order.quantity, ib_order_price)
				ibOrder.account = managed_trade.template.account
				ibOrder.outsideRth = True
			elif order.type == OrderType.STOP:
				ibOrder = StopOrder(self._mappedOrderAction(order.action), order.quantity, ib_order_price)
				ibOrder.account = managed_trade.template.account
				ibOrder.outsideRth = True
			elif order.type == OrderType.MARKET:
				ibOrder = MarketOrder(self._mappedOrderAction(order.action), order.quantity)
				ibOrder.account = managed_trade.template.account
				ibOrder.outsideRth = True
				ibOrder.auxPrice = 0
				ibOrder.lmtPrice = 0
			else:
				logger.error(f'Order type {order.type} currently not supported by IBKR Connector!')
			ibOrder.orderRef = order.orderReference
			if order.ocaGroup:  # https://interactivebrokers.github.io/tws-api/oca.html
				ibOrder.ocaGroup = order.ocaGroup
				ibOrder.ocaType = 1
			trade = self._ib.placeOrder(comboContract, ibOrder)
			order.brokerSpecific['comboContract'] = comboContract
			order.brokerSpecific['order'] = ibOrder
			order.brokerSpecific['trade'] = trade
			self._orders.append(order)
			logger.debug(f'Account: {managed_trade.template.account} Order {ibOrder.orderId} placed for Trade: {trade} Number of contracts: {order.quantity}')

	async def place_complex_order(self, take_profit_order: GenericOrder, stop_loss_order: GenericOrder, template: Template) -> bool:
		"""
		TWS doesn't support complex orders
		"""
		raise NotImplementedError()
	
	async def adjustOrder(self, managed_trade: ManagedTrade, order: GenericOrder, price: float) -> bool:
		""" 
		Adjusts the given order with the given new price
		"""
		if order.status == GenericOrderStatus.FILLED:
			logger.info('Order {} is already filled. Adjustment not required.', order)
			return True
		if order.status == GenericOrderStatus.CANCELLED:
			logger.info('Order {} is already cancelled. Adjustment not possible.', order)
			return True

		order.price = price
		ib_order_price = self._determine_ib_order_price(order)

		trade: Trade = order.brokerSpecific['trade']
		try:
			comboContract = order.brokerSpecific['comboContract']
			ibOrder: Order = order.brokerSpecific['order']
			if order.type == OrderType.LIMIT:
				ibOrder.lmtPrice = ib_order_price
			elif order.type == OrderType.STOP:
				ibOrder.auxPrice = ib_order_price
			trade = self._ib.placeOrder(comboContract, ibOrder)
			order.brokerSpecific['trade'] = trade
			return True
		except Exception as excp:
			logger.error('Exception beim Anpassen der Order')
		return False

	async def requestTickerData(self, symbols: List[str]) -> None:
		""" 
		Request ticker data for the given symbols and their options
		"""
		retries = 0
		qualifiedContracts = False
		contracts = []
		for symbol in symbols:
			requestOptionsData = False
			match symbol:
				case 'SPX':
					symbolData = IBSymbolData()
					symbolData.symbol = symbol
					symbolInformation = symbolInfo.symbol_infos[symbolData.symbol]
					symbolData.contract = Index('SPX', 'CBOE')
					symbolData.option_trading_class = symbolInformation.trading_class
					self._symbolData[symbol] = symbolData
					self._symbolReverseLookup[symbolData.contract.symbol] = symbol
					contracts.append(symbolData.contract)
				case 'VIX':
					symbolData = IBSymbolData()
					symbolData.symbol = symbol
					symbolData.contract = Index('VIX', 'CBOE')
					symbolData.trade_options = False  # No options for VIX
					self._symbolData[symbol] = symbolData
					self._symbolReverseLookup[symbolData.contract.symbol] = symbol
					contracts.append(symbolData.contract)
				case _:
					logger.error(f'Symbol {symbol} currently not supported by IBKR Connector!')
					continue

		while retries < 3:
			try:
				qualifiedContracts = await asyncio.wait_for(self._ib.qualifyContractsAsync(*contracts), timeout=10)
				for contract in qualifiedContracts:
					if self._symbolData[contract.symbol].trade_options:
						symbolInformation = symbolInfo.symbol_infos[contract.symbol]
						chains = await self._ib.reqSecDefOptParamsAsync(contract.symbol, '', contract.secType, contract.conId)
						if len(chains) > 0:
							chain = next(c for c in chains if c.tradingClass == symbolInformation.trading_class and c.exchange == contract.exchange)
							if chain == None:
								logger.error(f'No Option Chain for trading class {symbolInformation.trading_class} found. Unable to request option data')

							current_date = dt.date.today()
							expiration = current_date.strftime('%Y%m%d')
							symbolData = self._symbolData[contract.symbol]

							# Keep Option Chain expirations for later DTE calculations
							for expiration in chain.expirations:
								exp_date = dt.datetime.strptime(expiration, '%Y%m%d').date()
								if not exp_date in symbolData.expirations:
									symbolData.expirations.append(exp_date)

							symbolData.strikes = chain.strikes
							#if int(chain.expirations[0]) > int(expiration):
							if current_date > symbolData.expirations[0]:
								logger.warning(f'There are no {contract.symbol} options expiring today!')
					
					self._ib.reqMktData(contract, '', False, False)
				contracts_qualified = True
				break
			except asyncio.TimeoutError:
				retries += 1
				logger.warning(f'Timeout while qualifying contracts. Retrying ({retries}) in 5 seconds ...')
				await asyncio.sleep(5)
			except Exception as excp:
				logger.error(f'Error qualifying contract {symbolData.contract}: {excp}')
			
		if qualifiedContracts == False:
			logger.error(f'Error qualifying contracts when requesting ticker data!!')
		else:
			self._subscribed_ticker_data = True # Remember that we have subscribed to ticker data

	async def _connect_tws_task(self):
		try:
			await self._ib.connectAsync(self._host, self._port, clientId=self._clientid)
			
			# OTB-256: Check TWS version for SEC rule update (RIN 3235-AN23)
			await self._check_tws_version()
			
			self._emitConnectedEvent()
			
		except Exception as excp:
			self._emitConnectFailedEvent()
			#logger.error("Error connecting TWS: {}", excp)
			#attempt += 1
			#logger.error('Connect failed. Retrying {}. attempt in {} seconds', attempt, delaySecs)
			#await asyncio.sleep(delaySecs)
			#asyncio.create_task(self._connect_ib_task(attempt, delaySecs))

	async def _check_tws_version(self):
		"""
		OTB-256: Check TWS version for SEC rule update (RIN 3235-AN23)
		
		Due to SEC rule update, TWS must be at least version 10.37 
		starting November 3, 2025.
		
		Note: The TWS API only provides the API protocol version (serverVersion),
		not the actual TWS version number. We use the protocol version as a proxy:
		- API version 176+ generally corresponds to TWS 10.37+
		
		Behavior:
		- Before Nov 3, 2025: Warning only, continue operation
		- From Nov 3, 2025 onwards: Error and stop OptraBot
		"""
		import sys
		from datetime import date
		
		# Enforcement date for SEC rule
		ENFORCEMENT_DATE = date(2025, 11, 3)
		MIN_REQUIRED_API_VERSION = 176  # API version that corresponds to TWS 10.37+
		
		try:
			# Get TWS API protocol version from connection
			# Note: This is NOT the TWS version number (10.37), but the API protocol version
			api_version = self._ib.client.serverVersion()
			
			logger.info(f'Connected to TWS - API Protocol Version: {api_version} (checking for SEC compliance)')
			
			# Check if API version meets requirement
			# Note: TWS 10.37+ typically uses API version 176+
			# Example: TWS 10.37.1l uses API version 178, which is >= 176 → OK
			if api_version < MIN_REQUIRED_API_VERSION:
				today = date.today()
				
				if today >= ENFORCEMENT_DATE:
					# After enforcement date: Critical error, stop OptraBot
					error_msg = (
						f'❌ CRITICAL: TWS API version {api_version} is too old!\n'
						f'Due to SEC rule update (RIN 3235-AN23), TWS version 10.37 or newer is required since {ENFORCEMENT_DATE}.\n'
						f'Your API version {api_version} indicates an older TWS version.\n'
						f'Please upgrade your TWS to version 10.37 or newer.\n'
						f'OptraBot will now stop to prevent trading issues.'
					)
					logger.error(error_msg)
					
					# Send notification
					from optrabot.tradinghubclient import TradinghubClient, NotificationType
					try:
						await TradinghubClient().send_notification(
							NotificationType.ERROR,
							f'🚨 *CRITICAL: TWS Version Outdated*\n\n'
							f'API Protocol Version: {api_version}\n'
							f'Required: TWS 10.37+ (API version {MIN_REQUIRED_API_VERSION}+)\n\n'
							f'SEC rule RIN 3235-AN23 enforcement active since {ENFORCEMENT_DATE}.\n'
							f'*OptraBot stopped to prevent trading issues.*'
						)
					except Exception as notify_error:
						logger.error(f'Failed to send notification: {notify_error}')
					
					# Disconnect and stop
					self._ib.disconnect()
					sys.exit(1)
				else:
					# Before enforcement date: Warning only
					days_until_enforcement = (ENFORCEMENT_DATE - today).days
					warning_msg = (
						f'⚠️ WARNING: TWS API version {api_version} may be too old!\n'
						f'TWS version 10.37 or newer (API version {MIN_REQUIRED_API_VERSION}+) will be required starting {ENFORCEMENT_DATE} '
						f'({days_until_enforcement} days remaining).\n'
						f'Your current API version {api_version} indicates an older TWS version.\n'
						f'Please upgrade your TWS soon to avoid disruption.\n'
						f'SEC rule update: RIN 3235-AN23 - Round lot size changes for US stocks.'
					)
					logger.warning(warning_msg)
					
					# Send notification
					from optrabot.tradinghubclient import TradinghubClient, NotificationType
					try:
						await TradinghubClient().send_notification(
							NotificationType.WARN,
							f'⚠️ *TWS Version Warning*\n\n'
							f'API Protocol Version: {api_version}\n'
							f'Required from {ENFORCEMENT_DATE}: TWS 10.37+ (API version {MIN_REQUIRED_API_VERSION}+)\n\n'
							f'*{days_until_enforcement} days remaining until enforcement.*\n'
							f'Please upgrade TWS to 10.37+ soon (SEC rule RIN 3235-AN23).'
						)
					except Exception as notify_error:
						logger.error(f'Failed to send notification: {notify_error}')
			else:
				logger.success(f'✅ TWS API version {api_version} meets requirements (TWS 10.37+ uses API version {MIN_REQUIRED_API_VERSION}+)')
				
		except Exception as e:
			logger.error(f'Failed to check TWS version: {e}')

	async def onDisconnected(self):
		#logger.warning('Disconnected from TWS, attempting to reconnect in 30 seconds ...')
		await self.set_trading_enabled(False, "Broker disconnected")
		self._emitDisconnectedEvent()

	async def onErrorEvent(self, reqId: int, errorCode: int, errorString: str, contract: Contract):
		if errorCode in {104, 200, 202, 399, 2103, 2104, 2105, 2106, 2107, 2108, 2109, 2157, 2158}:
			# 104: Cannot modify a filled order
			# 200: No security definition has been found for the request, contract
			# 202: Order wurde storniert z.B. TP order bei OCO, wenn SL Order ausgeführt wurde
			# 399: Warnung das die Order nur während der regulären Handelszeiten ausgeführt wird
			# 10147: OrderId ... that needs to be cancelled is not found
			# 2103, 2104, 2105, 2106, 2108, 2158: Marktdatenverbindung ignorieren
			# 2107: Die Verbindung zum HMDS-Datenzentrum is inaktv, sollte jedoch auf Anfrage verfügbar seind
			# 2109: Warnhinweis zu einem Orderereignis außerhalb der regulären Handelszeiten
			# 2157: Verbindung zum Sec-def-Datenzentrum unterbrochen
			return
		elif errorCode == 201:
			# 201: Order rejected - reason:Stop price revision is disallowed after order has triggered
			# 201: Order rejected - reason:We are unable to accept your order. Your Available Funds are in sufficient
			# 201: Order abgewiesen - Grund:Order has been cancelled already, too late to replace
			if errorString.find('Margin') > -1:
				logger.error('Order rejected: Insufficient margin!')
				# Send Notification
				from optrabot.tradinghubclient import TradinghubClient, NotificationType
				notification_message = f'‼️ Broker {self.broker}: Order rejected because of insufficient margin!'
				await TradinghubClient().send_notification(NotificationType.INFO, notification_message)
				return

		elif errorCode == 10147:
			# Die zu stornierende Order wurde nicht gefunden
			if reqId in self._duplicate_order_ids:
				logger.info(f'Order {reqId}' + ' is a duplicate order ID. Ignoring error event.')
				return

		elif errorCode == 10148:
			# Order that should be cancelled cannot be cancelled
			# Passiert beim Cancel einer TP oder SL Order, wenn sie per OCO Group verbunden sind und eine der Orders bereits
			# storniert wurde

			# Prüfen ob die betreffende Order in den conflicting Orders enthalten ist
			for entry in self._conflictingOrderMap.keys():
				order: GenericOrder = entry
				if order.brokerSpecific['trade'].order.orderId == reqId:
					logger.debug(f'Order {reqId} is conflicting with another order. Ignoring error event.')
					return
		
		elif errorCode == 10168:
			# Für die angeforderten Markdaten besteht kein Abonnement
			logger.error(f'No valid market data subscription for symbol {contract.symbol}')
			await self.set_trading_enabled(False, "No valid market data subscription")
			return

		elif errorCode == 1100:
			# Connection between TWS and IB lost.
			logger.warning('Connection between TWS and Interactive Brokers lost -> Trading disabled!')
			await self.set_trading_enabled(False, "Connection lost")
			return
		elif errorCode == 1102:
			# Connection between TWS and IB restored
			logger.success('Connection between TWS and Interactive Brokers has been reestablished! -> Trading enabled!')
			await self.set_trading_enabled(True, "Connection restored")
			return
		elif errorCode == 321:
			# Invalid contract id - happens after overnight disconnection
			# This will be handled by the reconnection flow which triggers requestTickerData()
			logger.warning('Invalid contract id detected - will be refreshed on next reconnection')
			return
		elif errorCode == 2148:
			# Margin Warning
			logger.warning('IB reports potential margin violation!')
			return

		errorData = {'action': 'errorEvent','reqId':reqId, 'errorCode':errorCode, 'errorString':errorString, 'contract': contract}
		logger.error('IB raised following error: {}', errorData)
	
	async def onExecDetailsEvent(self, trade: Trade, fill: Fill):
		""" This eventhandler is called on trade execution.
		The event is raised for each single fill of an order leg
		"""
		logger.debug('Exec Detail for trade {}', trade)
		logger.debug('Fill: {}', fill)
		relevantOrder = None
		for order in self._orders:
			brokerSpecificTrade: Trade = order.brokerSpecific['trade']
			if brokerSpecificTrade.order.orderId == trade.order.orderId:
				relevantOrder = order
				break
		if relevantOrder == None:
			logger.debug(f'No managed order matched the execution details event')
			return
		if fill.contract.secType == 'BAG':
			# Do not store inform stion on combo executions
			return
		if fill.execution.side == 'BOT':
			action = OrderAction.BUY
		else:
			action = OrderAction.SELL
		expirationDate = dt.datetime.strptime(fill.contract.lastTradeDateOrContractMonth, '%Y%m%d')
		execution = Execution(id=fill.execution.execId, 
							action=action,
							sec_type=fill.contract.right,
							strike=fill.contract.strike,
							amount=fill.execution.shares,
							price=fill.execution.avgPrice,
							expiration=expirationDate,
							timestamp=fill.execution.time)
		self._emitOrderExecutionDetailsEvent(relevantOrder, execution)

	async def onCommissionReportEvent(self, trade: Trade, fill: Fill, report: CommissionReport):
		"""
		Handles the Commission Report Event
		"""
		relevantOrder = None
		logger.debug('Commission Report order id {} and fill {}', trade.order.orderId, fill)
		logger.debug('Commission Report: {}', report)
		for order in self._orders:
			brokerSpecificTrade: Trade = order.brokerSpecific['trade']
			if brokerSpecificTrade.order.orderId == trade.order.orderId:
				relevantOrder = order
				break
		if relevantOrder == None:
			logger.debug(f'No managed order matched the commission report event')
			return

		self._emitCommissionReportEvent(relevantOrder, fill.execution.execId, report.commission, 0)

	async def onOrderStatusEvent(self, trade: Trade):
		"""
		Handles the Order Status Event
		"""
		logger.debug(f'onOrderStatusEvent() Order: {trade.order.orderId} Status: {trade.orderStatus.status}')
		relevantOrder: GenericOrder = None
		for order in self._orders:
			brokerSpecificTrade: Trade = order.brokerSpecific['trade']
			if brokerSpecificTrade.order.orderId == trade.order.orderId:
				relevantOrder = order
				break
		if relevantOrder == None:
			logger.debug(f'No managed order matched the status event')
			return
		
		# Prüfen ob die order in der Conflicting Order Map enthalten ist
		if relevantOrder in self._conflictingOrderMap.keys():
			causingOrder = self._conflictingOrderMap[relevantOrder]
			causingOrderOrderId = causingOrder.brokerSpecific['trade'].order.orderId
			logger.debug(f'Order {trade.order.orderId} is conflicting with order {causingOrderOrderId}. Ignoring status event.')
			return

		if trade.orderStatus.status == OrderStatus.Cancelled or trade.orderStatus.status == OrderStatus.Inactive or trade.orderStatus.status == OrderStatus.ApiCancelled:
			log_entry: TradeLogEntry = None
			for log_entry in trade.log:
				if log_entry.status != 'Cancelled':
					continue
				logger.debug(f'Order Log: {log_entry}')
			if log_entry == None:
				logger.error(f'No log entry found for cancelled order!')
			elif log_entry.errorCode == 103:
				# Error 103, reqId 10292: Doppelt vorhandene Order-ID
				logger.warning('Adjustment of entry order has been rejected, because Duplicate Order-ID.')
				duplicate_order_id = trade.order.orderId
				self._duplicate_order_ids.append(duplicate_order_id)
				logger.debug(f'Added Order ID {duplicate_order_id} to duplicate order IDs')
				# In diesem Fall muss die Order sicherheitshalber neu platziert werden
				try:
					self._ib.cancelOrder(trade.order)
				except Exception as excp:
					logger.debug(f'Error cancelling order: {excp}')
				new_order_id = self._ib.client.getReqId()
				trade.order.orderId = new_order_id
				# Add the replacement order ID to duplicate list to prevent it from being treated as cancelled
				self._duplicate_order_ids.append(new_order_id)
				logger.debug(f'Replacing order with new Order Id {trade.order.orderId}')
				relevantOrder.brokerSpecific['trade'] = self._ib.placeOrder(trade.contract, trade.order)
				logger.info(f'Order {duplicate_order_id} has been replaced with new Order Id {trade.order.orderId}')
				return
			elif log_entry.errorCode == 104:
				# Error 104: Cannot modify a filled order
				logger.warning(f'Order {trade.order.orderId} has been filled already. Ignoring the Cancelled status event.')
				return
			
			# Ingnore the Cancelled status event for duplicate order IDs
			if trade.order.orderId in self._duplicate_order_ids:
				logger.debug(f'Order {trade.order.orderId} is a duplicate order ID. Ignoring the Cancelled status event.')
				return
			else:
				logger.debug(f'Order with ID {trade.order.orderId} is not in the list of duplicate order IDs')

			# Wenn die stornierte Order eine verursachende Order für Conflicting Ordrer(s) ist, dann müssen die stornierten Orders wieder in den Markt gelegt werden
			await self._placeConflictingOrdersAgain(relevantOrder)
		
		filledAmount = 0
		if trade.orderStatus.status == OrderStatus.Filled:
			# Wenn die gefüllte Order eine verursachende Order für conflicting Order(s) ist und sie vollständig ausgeüfhrt wurde,
			# dann müssen die stornierten Orders wieder in den Markt gelegt werden
			await self._placeConflictingOrdersAgain(relevantOrder)

			filledAmount = relevantOrder.quantity - trade.remaining()
			relevantOrder.averageFillPrice = trade.orderStatus.avgFillPrice
			
			# Remove filled order from duplicate order IDs list (if it's a replacement order)
			if trade.order.orderId in self._duplicate_order_ids:
				self._duplicate_order_ids.remove(trade.order.orderId)
				logger.debug(f'Removed filled order {trade.order.orderId} from duplicate order IDs list')

		self._emitOrderStatusEvent(relevantOrder, self._genericOrderStatus(trade.orderStatus.status), filledAmount)

	def getFillPrice(self, order: GenericOrder) -> float:
		""" 
		Returns the fill price of the given order if it is filled
		"""
		trade: Trade = order.brokerSpecific['trade']
		return abs(trade.orderStatus.avgFillPrice)
	
	def getLastPrice(self, symbol: str) -> float:
		""" 
		Returns the last price of the given symbol
		"""
		symbolData = self._symbolData[symbol]
		return symbolData.lastPrice

	def get_option_strike_data(self, symbol: str, expiration: dt.date) -> OptionStrikeData:
		""" 
		Returns the option strike data for the given symbol and expiration. It is including
		prices and greeks.
		"""
		symbolData = self._symbolData[symbol]
		try:
			return symbolData.optionPriceData[expiration]
		except KeyError:
			raise ValueError(f'No option strike data for symbol {symbol} and expiration {expiration} found!')

	def get_option_strike_price_data(self, symbol: str, expiration: dt.date, strike: float) -> OptionStrikePriceData:
		""" 
		Returns the option strike price data for the given symbol, expiration date, strike price and right
		"""
		symbolData = self._symbolData[symbol]
		try:
			optionStrikeData = symbolData.optionPriceData[expiration]
			if strike in optionStrikeData.strikeData.keys():
				return optionStrikeData.strikeData[strike]
			else:
				return None
		except KeyError as keyErr:
			return None

	def get_strike_by_delta(self, symbol: str, right: str, delta: int) -> float:
		""" 
		Returns the strike price based on the given delta based on the buffered option price data
		"""
		symbolData = self._symbolData[symbol]
		current_date = dt.date.today()
		option_price_data: OptionStrikeData= symbolData.optionPriceData[current_date]
		previous_delta = 0
		previous_strike = 0
		reverse = True if right == OptionRight.PUT else False
		sorted_strikes = dict(sorted(option_price_data.strikeData.items(), reverse=reverse))
		for strike, price_data in sorted_strikes.items():
			if right == OptionRight.PUT:
				if price_data.putDelta == None:
					continue
				adjusted_delta = price_data.putDelta * -100
			else:
				if price_data.callDelta == None:
					continue
				adjusted_delta = price_data.callDelta * 100
			if adjusted_delta <= delta:
				if OptionHelper.closest_number(delta, previous_delta, adjusted_delta) == adjusted_delta:
					return strike
				else:
					return previous_strike
			previous_delta = adjusted_delta
			previous_strike = strike

		raise ValueError(f'No strike price found for delta {delta} in symbol {symbol}!')
	
	def get_strike_by_price(self, symbol: str, right: str, price: float) -> float:
		""" 
		Returns the strike price based on the given premium price based on the buffered option price data
		"""
		symbolData = self._symbolData[symbol]
		current_date = dt.date.today()
		option_price_data: OptionStrikeData= symbolData.optionPriceData[current_date]
		previous_price = 0
		previous_strike = 0
		reverse = True if right == OptionRight.PUT else False
		sorted_strikes = dict(sorted(option_price_data.strikeData.items(), reverse=reverse))
		for strike, price_data in sorted_strikes.items():
			if right == OptionRight.PUT:
				current_strike_price = price_data.getPutMidPrice()
				if current_strike_price == None:
					continue
			elif right == OptionRight.CALL:
				current_strike_price = price_data.getCallMidPrice()
				if current_strike_price == None:
					continue
			if current_strike_price > 0 and current_strike_price <= price:
				if OptionHelper.closest_number(price, previous_price, current_strike_price) == current_strike_price:
					return strike
				else:
					return previous_strike
			previous_price = current_strike_price
			previous_strike = strike
		raise ValueError(f'No strike price found for price {price} in symbol {symbol}!')

	async def onPendingTickers(self, tickers: List[Ticker]):
		"""
		Handles the pending tickers event
		"""
		if not self._subscribed_ticker_data:
			return
		
		for ticker in tickers:
			if ticker.contract.symbol in self._symbolData.keys():
				logger.trace(f'Ticker for symbol {ticker.contract.symbol} received.')
				ticker.lastExchange
				logger.trace(f'Ticker Last {ticker.last}: {ticker}')
				symbolData = self._symbolData[ticker.contract.symbol]
				symbolData.ticker = ticker
				if ticker.contract.secType != 'OPT':
					# Capture last price of Underlying
					if ticker.last and util.isNan(ticker.last) == False:
						symbolData.lastPrice = ticker.last
					else:
						symbolData.lastPrice = ticker.close

				# Prüfen ob Preisdaten vorhanden sind
				if util.isNan(symbolData.lastPrice) and ticker.contract.secType != 'OPT':
					# Wenn mehrmals keine gültigen Preisdaten empfangen wurden, dann wird der Handel deaktiviert
					symbolData.noPriceDataCount += 1
					if symbolData.noPriceDataCount > 5:
						logger.error(f'Receiving no valid price data for symbol {symbolData.symbol}. Trading disabled!')
						await self.set_trading_enabled(False, "No valid price data")
						break
					continue
				else:
					symbolData.noPriceDataCount = 0
					if self.isTradingEnabled() == False:
						await self.set_trading_enabled(True, "Price valid data received")

				# Für VIX werden keine weiteren Daten gelesen
				if ticker.contract.symbol == 'VIX':
					break

				if ticker.contract.secType != 'OPT':
					# Keine Optionen -> Fehlende Optionen abfragen
					try:
						atmStrike = OptionHelper.roundToStrikePrice(symbolData.lastPrice)
						if symbolData.lastAtmStrike != atmStrike:  # Check for missing Option Data only if ATM Strike has changed
							
							symbolData.lastAtmStrike = atmStrike
							await self._requestMissingOptionData(symbolData, atmStrike)
					except KeyError as keyErr:
						logger.error(f'No generic symbol found for IB TWS symbol {ticker.contract.symbol}')
				else:
					# Option verarbeiten
					try:
						optionType, expiration, strike = self._getOptionInfos(ticker.contract)
					except Exception as exc:
						logger.error(f'Error getting option infos: {exc}')
					genericSymbol = self._symbolReverseLookup[ticker.contract.symbol]
					symbol_information = symbolInfo.symbol_infos[genericSymbol]
					optionStrikePriceData = self.get_option_strike_price_data(genericSymbol, expiration, strike)
					if optionStrikePriceData:
						if optionType == OptionRight.CALL:
							optionStrikePriceData.callAsk = ticker.ask
							optionStrikePriceData.callBid = ticker.bid
							if ticker.modelGreeks:
								optionStrikePriceData.callDelta = ticker.modelGreeks.delta
						else:
							optionStrikePriceData.putAsk = ticker.ask
							optionStrikePriceData.putBid = ticker.bid
							if ticker.modelGreeks:
								optionStrikePriceData.putDelta = ticker.modelGreeks.delta
						optionStrikePriceData.lastUpdated = dt.datetime.now(symbol_information.timezone)
						self._last_option_price_update_time = optionStrikePriceData.lastUpdated
					else:
						logger.debug(f'No option data available for strike {strike} in symbol {genericSymbol} yet. Ignoring Ticker data.')
					
	def uses_oco_orders(self) -> bool:
		""" 
		The TWS Connector uses OCO orders for take profit and stop loss orders
		"""
		return True

	async def unsubscribe_ticker_data(self):
		""" 
		Unsubscribes all ticker data subscriptions
		"""
		await super().unsubscribe_ticker_data()
		for symbolData in self._symbolData.values():
			try:
				for day_option_price_data in symbolData.optionPriceData.values():
					for value in day_option_price_data.strikeData.values():
						option_price_data: OptionStrikePriceData = value
						try:
							call_contract = option_price_data.brokerSpecific['call_contract']
							if call_contract:
								self._ib.cancelMktData(call_contract)
								option_price_data.brokerSpecific['call_contract'] = None
						except KeyError as keyErr:
							pass
						try:
							put_contract = option_price_data.brokerSpecific['put_contract']
							if put_contract:
								self._ib.cancelMktData(put_contract)
								option_price_data.brokerSpecific['put_contract'] = None
						except KeyError as keyErr:
							pass
				symbolData.optionPriceData.clear()
				symbolData.lastAtmStrike = 0
				symbolData.lastPrice = 0
			except KeyError as keyErr:
				pass
	
	def _createReplacementOrder(self, ibOrder: Order) -> Order:
		"""
		Creates a new replacement order for the given order
		"""
		if isinstance(ibOrder, LimitOrder):
			replacementOrder = LimitOrder(ibOrder.action, ibOrder.totalQuantity, ibOrder.lmtPrice)
		elif isinstance(ibOrder, StopOrder):
			replacementOrder = StopOrder(ibOrder.action, ibOrder.totalQuantity, ibOrder.auxPrice)
		else:
			logger.error(f'Order type {type(ibOrder)} currently not supported by IBKR Connector!')
			return None
		replacementOrder.account = ibOrder.account
		replacementOrder.outsideRth = ibOrder.outsideRth
		replacementOrder.orderRef = ibOrder.orderRef
		replacementOrder.ocaGroup = ibOrder.ocaGroup
		replacementOrder.ocaType = ibOrder.ocaType
		replacementOrder.transmit = True
		return replacementOrder

	def _determine_ib_order_price(self, order: GenericOrder) -> float:
		"""
		Determines the correct IBKR order price based on the given generic order
		"""
		assert order.price_effect is not None

		# If it is a single leg order, and it is a buy to open or sell to close order, the price is positive
		if len(order.legs) == 1 and (order.action == OrderAction.BUY_TO_OPEN or order.action == OrderAction.SELL_TO_CLOSE):
			return abs(order.price)

		# Calculate price based on the order legs
		calculated_price = order.calculate_price()
		# If it is a closing order, the calculated price must be inverted, because the leg actions are inverted as well (_mapped_leg_action)
		if order.action == OrderAction.SELL_TO_CLOSE or order.action == OrderAction.BUY_TO_CLOSE:
			calculated_price = -calculated_price

		# Invert the order price if signs are different from calculated price
		ib_order_price = order.price
		if (calculated_price < 0 and order.price > 0) or (calculated_price > 0 and order.price < 0):
			ib_order_price = -order.price
		return ib_order_price

	def _findOrderByIBTrade(self, trade: Trade) -> GenericOrder:
		"""
		Find the generic order that belongs to the given IBKR trade
		"""
		for order in self._orders:
			if order.brokerSpecific['trade'] == trade:
				return order
		return None

	async def _findConflictingOrders(self, order: GenericOrder, account: str) -> List[Trade]:
		"""
		Find all open trades for the given account and symbol that are conflicting with the given order
		"""
		comboLegs = order.brokerSpecific['comboLegs']
		oppositeAction = 'BUY' if self._mappedOrderAction(order.action) == 'SELL' else 'SELL'
		conflictingTrades: List[Trade] = []
		openTrades: List[Trade] = await self._ib.reqAllOpenOrdersAsync()
		
		for openTrade in openTrades:
			if openTrade.order.transmit == False or openTrade.contract.symbol != order.symbol or openTrade.order.account != account or openTrade.order.action != oppositeAction:
				continue

			# ausgeführte und stonierte Trades/Orders spielen keine Rolle (sollten eigentlich nicht mehr in der Liste sein)
			if openTrade.orderStatus.status == OrderStatus.Cancelled or openTrade.orderStatus.status == OrderStatus.Filled:
				continue

			# Offene Order mit entgegengesetzter Aktion gefunden -> Sind enhalten sie Kontrakte die mit der neuen Order kollidieren
			if openTrade.contract.secType == 'BAG':
				for leg in openTrade.contract.comboLegs:
					for comboLeg in comboLegs:
						if leg.conId == comboLeg.conId:
							if openTrade not in conflictingTrades:
								conflictingTrades.append(openTrade)
			else:
				for comboLeg in comboLegs:
					if openTrade.contract.conId == comboLeg.conId:
						if openTrade not in conflictingTrades:
							conflictingTrades.append(openTrade)
		return conflictingTrades

	def _genericOrderStatus(self, status: OrderStatus) -> GenericOrderStatus:
		"""
		Maps the IBKR specific order status to the general order status
		"""
		match status:
			case OrderStatus.PendingSubmit:
				return GenericOrderStatus.OPEN
			case OrderStatus.PreSubmitted:
				return GenericOrderStatus.OPEN
			case OrderStatus.Submitted:
				return GenericOrderStatus.OPEN
			case OrderStatus.ApiCancelled:
				return GenericOrderStatus.CANCELLED
			case OrderStatus.Cancelled:
				return GenericOrderStatus.CANCELLED
			case OrderStatus.Filled:
				return GenericOrderStatus.FILLED
			case OrderStatus.Inactive:
				return GenericOrderStatus.CANCELLED
			case OrderStatus.PendingCancel:
				return GenericOrderStatus.OPEN
			case OrderStatus.Cancelled:
				return GenericOrderStatus.CANCELLED

	def _getOptionInfos(self, option: Option) -> tuple:
		"""
		Extracts the generic symbol and expiration date, strike price from the IBKR option contract
		"""
		error = False
		match = self._compiled_option_pattern.match(option.localSymbol)
		try:
			if match:
				for symbol, symbol_info in symbolInfo.symbol_infos.items():
					if symbol_info.symbol == option.symbol:
						genericSymbol = symbol_info.symbol
						break
				expiration = dt.datetime.strptime(match.group('expiration'), '%y%m%d').date()
				optionType = OptionRight.CALL if match.group('type') == 'C' else OptionRight.PUT
			else:
				raise IndexError
		except IndexError as indexErr:
			logger.error(f'Invalid option symbol {option.localSymbol}')
			error = True
		except ValueError as valueErr:
			logger.error(f'Invalid option symbol {option.localSymbol}')
			error = True
		if genericSymbol == None or error == True:
			raise ValueError(f'Invalid option symbol {option.localSymbol}')
		return optionType, expiration, option.strike

	def _mappedOrderAction(self, action: OrderAction) -> str:
		"""
		Maps the general order action to the IBKR specific order action
		"""
		match action:
			case OrderAction.BUY:
				return 'BUY'
			case OrderAction.BUY_TO_OPEN:
				return 'BUY'
			case OrderAction.BUY_TO_CLOSE:
				return 'BUY'
			case OrderAction.SELL:
				return 'SELL'
			case OrderAction.SELL_TO_OPEN:
				return 'SELL'
			case OrderAction.SELL_TO_CLOSE:
				return 'SELL'
			case _:
				logger.error(f'Order action {action} not supported by IBKR Connector!')
				return None
			
	def _mapped_leg_action(self, order_action: OrderAction, leg_action: OrderAction) -> str:
		"""
		Maps the leg action of a generic order to the correct IB specific leg action depending on the order action
		"""
		match order_action:
			case OrderAction.BUY_TO_OPEN:
				if leg_action == OrderAction.BUY:
					return 'BUY'
				elif leg_action == OrderAction.SELL:
					return 'SELL'
				else:
					raise ValueError(f'Unknown leg action: {leg_action}')
			case OrderAction.SELL_TO_OPEN:
				if leg_action == OrderAction.BUY:
					return 'BUY'
				elif leg_action == OrderAction.SELL:
					return 'SELL'
				else:
					raise ValueError(f'Unknown leg action: {leg_action}')
			case OrderAction.BUY_TO_CLOSE:
				# Action of the leg needs to be inverted
				if leg_action == OrderAction.BUY:
					return 'SELL'
				elif leg_action == OrderAction.SELL:
					return 'BUY'
				else:
					raise ValueError(f'Unknown leg action: {leg_action}')
			case OrderAction.SELL_TO_CLOSE:
				# Action of the leg needs to be inverted
				if leg_action == OrderAction.BUY:
					return 'SELL'
				elif leg_action == OrderAction.SELL:
					return 'BUY'
				else:
					raise ValueError(f'Unknown leg action: {leg_action}')
			case _:
				logger.error(f'Order action {order_action} not supported by IBKR Connector!')
				return None
	
	def _mappedRight(self, right: OptionRight) -> str:
		"""
		Maps the general option right to the IBKR specific option right
		"""
		match right:
			case OptionRight.CALL:
				return 'C'
			case OptionRight.PUT:
				return 'P'
			case _:
				raise ValueError(f'Option right {right} not supported by IBKR Connector!')
				return None
			
	async def _placeConflictingOrdersAgain(self, order: GenericOrder):
		"""
		Relevante temporär stornierte Conflicting Orders wieder in den Markt legen, wenn
		die übergebene Order eine verursachende Order ist
		"""
		resolvedConflictingOrders: List[GenericOrder] = []
		for key, value in self._conflictingOrderMap.items():
			if value == order:
				logger.debug(f'Causing order of a conflicting order has been filled. Re-placing conflicting order again.')
				if type(key) == ForeignOrder:
					logger.debug(f'Conflicting order is a foreign order. Re-placing it.')
				conflictingOrder: GenericOrder = key
				ibOrder:Order = conflictingOrder.brokerSpecific['order']
				replacementOrder = self._createReplacementOrder(ibOrder)
				replacementTrade = self._ib.placeOrder(conflictingOrder.brokerSpecific['comboContract'], replacementOrder)
				logger.debug(f'Conflicting order {ibOrder.orderId} has been re-placed with order id {replacementTrade.order.orderId}')
				conflictingOrder.brokerSpecific['order'] = replacementOrder
				conflictingOrder.brokerSpecific['trade'] = replacementTrade
				resolvedConflictingOrders.append(conflictingOrder)
		
		# Jetzt wiedereingestellte conflicting orders aus der Map entfernen
		for entry in resolvedConflictingOrders:
			self._conflictingOrderMap.pop(entry)

	async def _getAllOpenOrderContracts(self, account, symbol) -> List:
		"""
		Determine the ContractIds of all open Orders for the given account and symbol
		"""
		openTrades: List[Trade] = await self._ib.reqAllOpenOrdersAsync()
		openOrderContracts = list()
		for openTrade in openTrades:
			if openTrade.contract.symbol != symbol or openTrade.order.account != account:
				continue
			if openTrade.contract.secType == 'BAG':
				for leg in openTrade.contract.comboLegs:
					openOrderContracts.append(leg.conId)
			else:
				openOrderContracts.append(openTrade.contract.conId)
		return openOrderContracts
	
	def _calculateMidPrice(self, legs: List[Leg], tickers: List[Ticker], contracts: List[Contract]) -> float:
		"""
		Calculates the mid price for the given tickers
		"""
		midPrice = None
		for leg in legs:
			for ticker in tickers:
				if ticker.contract.strike == leg.strike and ticker.contract.right == self._mappedRight(leg.right):
					leg
					legMidPrice = (ticker.ask + ticker.bid) / 2
					if util.isNan(legMidPrice) or (ticker.ask == -1.00 and ticker.bid == -1.00):
						if leg.action == OrderAction.BUY:
							legMidPrice = 0.05
						else:
							return None
					if leg.action == OrderAction.SELL:
						midPrice = -legMidPrice
					else:
						midPrice += legMidPrice
					break

		return OptionHelper.roundToTickSize(midPrice)
	
	async def _requestMissingOptionData(self, symbolData: IBSymbolData, atmStrike: float):
		"""
		Request option data for the given symbol and expiration date
		"""
		relevant_expirations = self._determine_relevant_expirations(symbolData)
		symbolInformation = symbolInfo.symbol_infos[symbolData.symbol]
		strikes_of_interest = self._determine_strikes_of_interest(symbolData, atmStrike)	
		options_to_be_requested = []
		for relevant_expiration in relevant_expirations:
			# Bestehende gespeicherte Optionen abfragen
			try:
				expiration_date_str = relevant_expiration.strftime('%Y%m%d')
				optionStrikeData = symbolData.optionPriceData[relevant_expiration]
			except KeyError as keyErr:
				# Wenn noch keine Optionsdaten für das Verfallsdatum vorhanden sind
				# dann bei der TWS anfragen ob es Optionen gibt
				chains = await self._ib.reqSecDefOptParamsAsync(symbolData.contract.symbol, '', symbolData.contract.secType, symbolData.contract.conId)
				try:
					chain = next(c for c in chains if c.tradingClass == symbolData.option_trading_class and c.exchange == symbolData.contract.exchange)
					if int(chain.expirations[0]) > int(expiration_date_str):
						return
				
					optionStrikeData = OptionStrikeData()
					symbolData.optionPriceData[relevant_expiration] = optionStrikeData
			
				except Exception as excp:
					logger.error(f'Unexpected error: {excp}')
					return

			for strike_price in strikes_of_interest:
				strike_identifier = f"{expiration_date_str}#{strike_price}"
				if strike_identifier in self._strikes_wo_option:
					continue

				try:
					optionStrikeData.strikeData[strike_price]
				except KeyError as keyErr:
					option_strike_price_data = OptionStrikePriceData()
					call_option_to_be_requested = Option(symbolData.symbol , expiration_date_str, strike_price, self._mappedRight(OptionRight.CALL), symbolInformation.exchange , tradingClass=symbolData.option_trading_class)
					list_of_details = await self._ib.reqContractDetailsAsync(call_option_to_be_requested)
					if len(list_of_details) == 0:
						logger.info(f'No Options available at strike {strike_price}')
						self._strikes_wo_option.append(expiration_date_str + '#' + str(strike_price))
						continue

					optionStrikeData.strikeData[strike_price] = option_strike_price_data

					put_option_to_be_requested = Option(symbolData.symbol , expiration_date_str, strike_price, self._mappedRight(OptionRight.PUT), symbolInformation.exchange , tradingClass=symbolData.option_trading_class)

					option_strike_price_data.brokerSpecific['call_contract'] = call_option_to_be_requested
					options_to_be_requested.append(call_option_to_be_requested)
				
					option_strike_price_data.brokerSpecific['put_contract'] = put_option_to_be_requested
					options_to_be_requested.append(put_option_to_be_requested)
		
		if len(options_to_be_requested) > 0:
			retries = 0
			qualified_successfully = False
			while retries < 3:
				try:
					qualifiedContracts = await asyncio.wait_for(self._ib.qualifyContractsAsync(*options_to_be_requested), timeout=10)
					qualified_successfully = True
					break
				except asyncio.TimeoutError:
					retries += 1
					logger.warning(f'Timeout while qualifying contracts. Retrying ({retries}) in 5 seconds ...')
					await asyncio.sleep(5)
			
			if qualified_successfully == False:
				logger.error(f'Error qualifying contracts when requesting option data!!')
				return

			for qualified_contract in qualifiedContracts:
				self._ib.reqMktData(qualified_contract, '', False, False)



"""
OTB-258: NTP Time Synchronization Module

This module checks the system time against NTP servers to detect clock drift.
If the time difference exceeds 30 seconds, it sends a warning via Telegram.

The time check is scheduled to run:
- Once at startup
- Once daily at 6:00 AM Eastern Time
"""

import ntplib
from datetime import datetime, timezone
from typing import Optional, Tuple
from loguru import logger

from optrabot.tradinghubclient import TradinghubClient


class TimeSync:
    """
    Manages time synchronization monitoring with NTP servers.
    
    This class provides functionality to:
    - Query NTP servers for accurate time
    - Compare system time with NTP time
    - Send warnings via Telegram if drift exceeds threshold
    """
    
    # NTP servers in order of preference
    NTP_SERVERS = [
        'pool.ntp.org',       # Global NTP pool
        'time.google.com',    # Google's NTP server
        'time.apple.com'      # Apple's NTP server
    ]
    
    # Maximum allowed time difference in seconds
    MAX_DRIFT_SECONDS = 10
    
    def __init__(self, trading_hub_client: TradinghubClient =None):
        """
        Initialize TimeSync.
        
        Args:
            trading_hub_client: Optional TradingHubClient instance for sending alerts
        """
        self.trading_hub_client = trading_hub_client
        self.ntp_client = ntplib.NTPClient()
    
    @staticmethod
    def format_time_drift(seconds: float) -> str:
        """
        Format time drift in human-readable format.
        
        Automatically chooses the appropriate unit:
        - < 60s: seconds
        - < 3600s: minutes and seconds
        - < 86400s: hours and minutes
        - >= 86400s: days and hours
        
        Args:
            seconds: Time drift in seconds
            
        Returns:
            Formatted string (e.g., "45s", "2m 30s", "1h 15m", "2d 3h")
        """
        abs_seconds = abs(seconds)
        
        if abs_seconds < 60:
            return f"{abs_seconds:.1f}s"
        
        if abs_seconds < 3600:
            minutes = int(abs_seconds // 60)
            remaining_seconds = int(abs_seconds % 60)
            return f"{minutes}m {remaining_seconds}s"
        
        if abs_seconds < 86400:
            hours = int(abs_seconds // 3600)
            remaining_minutes = int((abs_seconds % 3600) // 60)
            return f"{hours}h {remaining_minutes}m"
        
        # >= 1 day
        days = int(abs_seconds // 86400)
        remaining_hours = int((abs_seconds % 86400) // 3600)
        return f"{days}d {remaining_hours}h"
    
    def get_ntp_time(self) -> Optional[datetime]:
        """
        Query NTP servers to get accurate current time.
        
        Tries multiple NTP servers in order until one responds.
        
        Returns:
            datetime: Current time from NTP server (UTC timezone-aware)
            None: If all NTP servers fail
        """
        for server in self.NTP_SERVERS:
            try:
                logger.debug(f"Querying NTP server: {server}")
                response = self.ntp_client.request(server, version=3, timeout=5)
                ntp_time = datetime.fromtimestamp(response.tx_time, tz=timezone.utc)
                logger.debug(f"NTP time from {server}: {ntp_time}")
                return ntp_time
            except Exception as e:
                logger.warning(f"Failed to get time from {server}: {e}")
                continue
        
        logger.error("All NTP servers failed")
        return None
    
    def get_system_time(self) -> datetime:
        """
        Get current system time.
        
        Returns:
            datetime: Current system time (UTC timezone-aware)
        """
        return datetime.now(timezone.utc)
    
    def calculate_time_drift(self) -> Optional[Tuple[float, datetime, datetime]]:
        """
        Calculate the time difference between system time and NTP time.
        
        Returns:
            Tuple[float, datetime, datetime]: (drift_seconds, system_time, ntp_time)
            None: If NTP time cannot be retrieved
        """
        ntp_time = self.get_ntp_time()
        if ntp_time is None:
            return None
        
        system_time = self.get_system_time()
        drift = (system_time - ntp_time).total_seconds()
        
        logger.debug(f"Time drift: {drift:.2f} seconds")
        logger.debug(f"System time: {system_time}")
        logger.debug(f"NTP time: {ntp_time}")
        
        return drift, system_time, ntp_time

    async def send_warning_notification(self, drift: float, system_time: datetime, ntp_time: datetime):
        """
        Send a warning about time drift.

        Args:
            drift: Time difference in seconds
            system_time: Current system time
            ntp_time: Correct NTP time
        """
        if self.trading_hub_client is None:
            logger.warning("Cannot send notification- no client configured")
            return
        
        try:
            from optrabot.tradinghubclient import NotificationType
            
            drift_formatted = self.format_time_drift(drift)
            
            message = (
                f"⚠️ TIME SYNC WARNING ⚠️\n\n"
                f"System time is off by {drift_formatted}.\n\n"
                f"System time: {system_time.strftime('%Y-%m-%d %H:%M:%S UTC')}\n"
                f"NTP time: {ntp_time.strftime('%Y-%m-%d %H:%M:%S UTC')}\n\n"
                f"Please check and adjust system time manually."
            )

            await self.trading_hub_client.send_notification(NotificationType.WARN, message)
            logger.debug("Telegram warning sent")
            
        except Exception as e:
            logger.error(f"Failed to send Telegram warning: {e}")
    
    async def check_and_sync_time(self):
        """
        Main time synchronization check.
        
        This method:
        1. Queries NTP servers for accurate time
        2. Calculates time drift
        3. Sends Telegram warning if drift > threshold
        
        This is the method called by APScheduler.
        """
        NUM_OF_CHARS = 70
        logger.debug("=" * NUM_OF_CHARS)
        logger.debug("Starting time synchronization check")
        logger.debug("=" * NUM_OF_CHARS)

        # Get time drift
        result = self.calculate_time_drift()
        if result is None:
            logger.error("Cannot check time drift - NTP servers unavailable")
            return
        
        drift, system_time, ntp_time = result
        
        # Check if drift exceeds threshold
        if abs(drift) <= self.MAX_DRIFT_SECONDS:
            drift_formatted = self.format_time_drift(drift)
            logger.info(f"✅ Time drift within acceptable range: {drift_formatted} (max: {self.MAX_DRIFT_SECONDS}s)")
            return
        
        # Drift exceeds threshold - send warning
        drift_formatted = self.format_time_drift(drift)
        logger.warning(
            f"⚠️ Time drift exceeds threshold: {drift_formatted} "
            f"(max: {self.MAX_DRIFT_SECONDS}s)"
        )

        # Send warning notification
        await self.send_warning_notification(drift, system_time, ntp_time)

        logger.debug("=" * NUM_OF_CHARS)
        logger.debug("Time synchronization check completed")
        logger.debug("=" * NUM_OF_CHARS)



import copy
from datetime import datetime
from typing import List
from loguru import logger
from optrabot.broker.order import Leg, OptionRight, Order, OrderAction, OrderType, PriceEffect
from optrabot.optionhelper import OptionHelper
from optrabot.signaldata import SignalData
from optrabot.tradetemplate.processor.templateprocessorbase import TemplateProcessorBase
from optrabot.tradetemplate.templatefactory import PutSpread, Template
from optrabot.managedtrade import ManagedTrade


class PutSpreadProcessor(TemplateProcessorBase):
	
	def __init__(self, template: Template):
		"""
		Initializes the put spread processor with the given template
		"""
		super().__init__(template)

	def composeEntryOrder(self, signalData: SignalData = None):
		"""
		Composes the entry order for the put spread template
		"""
		super().composeEntryOrder(signalData)
		putSpreadTemplate :PutSpread = self._template
		shortStrike = None
		# Short Strike Determination
		if signalData and signalData.strike > 0:
			shortStrike = signalData.strike
		else:
			shortStrikeData = putSpreadTemplate.getShortStrikeData()
			if not shortStrikeData:
				raise ValueError('Configuration for Short Strike is missing in template!')
			if shortStrikeData.offset:
				logger.debug(f'Using Short Strike Offset: {shortStrikeData.offset}')
				shortStrike =  OptionHelper.roundToStrikePrice(signalData.close + shortStrikeData.offset)
			if shortStrikeData.delta:
				logger.debug(f'Using Short Strike Delta: {shortStrikeData.delta}')
				try:
					shortStrike = self.get_short_strike_from_delta("SPX", OptionRight.PUT, shortStrikeData.delta)
				except ValueError as value_error:
					logger.error(f'Error while determining Short Strike by Delta: {value_error}')
					return None
			if shortStrikeData.price:
				logger.debug(f'Using Short Strike price: {shortStrikeData.price}')
				try:
					shortStrike = self.get_strike_by_price("SPX", OptionRight.PUT, shortStrikeData.price)
				except ValueError as value_error:
					logger.error(f'Error while determining Short Strike by Price: {value_error}')
					return None
		
		if shortStrike == None:
			raise ValueError('Short Strike could not be determined!')
			
		logger.debug(f'Using Short Strike: {shortStrike}')

		# Long Strike Determination
		longStrikeData = putSpreadTemplate.getLongStrikeData()
		if not longStrikeData:
			raise ValueError('Configuration for Long Strike is missing in template!')
		
		if longStrikeData.width:
			longStrike = OptionHelper.roundToStrikePrice(shortStrike - longStrikeData.width)
			logger.debug(f'Using Long Strike: {longStrike}')
		elif longStrikeData.price:
			logger.debug(f'Using Long Strike price: {longStrikeData.price}')
			longStrike = self.get_strike_by_price("SPX", OptionRight.PUT, longStrikeData.price)

		if longStrikeData.max_width:
			if longStrikeData.max_width < abs(shortStrike - longStrike):
				raise ValueError(f'Strikes {shortStrike}/{longStrike} exceeds the defined maximum width {longStrikeData.max_width}')

		# Now create the entry order with its legs as all required strikes are determined
		legs: List[Leg] = []
		legs.append(Leg(action=OrderAction.SELL, strike=shortStrike, symbol=self._template.symbol, right=OptionRight.PUT, expiration=self._determined_expiration_date, quantity=1))
		legs.append(Leg(action=OrderAction.BUY, strike=longStrike, symbol=self._template.symbol, right=OptionRight.PUT, expiration=self._determined_expiration_date, quantity=1))

		entryOrder = Order(symbol=self._template.symbol, legs=legs, action=OrderAction.BUY_TO_OPEN, quantity=self._template.amount, type=OrderType.LIMIT, price_effect=PriceEffect.CREDIT)
		return entryOrder
	
	def composeTakeProfitOrder(self, managedTrade: ManagedTrade, fillPrice: float) -> Order:
		"""
		Composes the take profit order based on the template and the given fill price
		"""
		super().composeTakeProfitOrder(managedTrade, fillPrice)
		logger.debug('Creating take profit order for template {}', self._template.name)
		takeProfitPrice = self._template.calculateTakeProfitPrice(fillPrice)
		logger.debug(f'Calculated take profit price: {takeProfitPrice}')
		tp_order_legs = copy.deepcopy(managedTrade.entryOrder.legs)
		self.invert_leg_actions(tp_order_legs)
		takeProfitOrder = Order(symbol=self._template.symbol, legs=tp_order_legs, action=OrderAction.SELL_TO_CLOSE, quantity=self._template.amount, type=OrderType.LIMIT, price=takeProfitPrice)
		return takeProfitOrder
	
	def composeStopLossOrder(self, managedTrade: ManagedTrade, fillPrice: float) -> Order:
		"""
		Composes the stop loss order based on the template and the given fill price
		"""
		super().composeStopLossOrder(managedTrade, fillPrice)
		logger.debug('Creating stop loss order for template {}', self._template.name)
		stopLossPrice = self._template.calculateStopLossPrice(fillPrice)
		logger.debug(f'Calculated stop loss price: {stopLossPrice}')
		sl_order_legs = copy.deepcopy(managedTrade.entryOrder.legs)
		self.invert_leg_actions(sl_order_legs)
		stopLossOrder = Order(symbol=self._template.symbol, legs=sl_order_legs, action=OrderAction.SELL_TO_CLOSE, quantity=self._template.amount, type=OrderType.STOP, price=stopLossPrice)
		return stopLossOrder


"""
Package for custom exceptions used throughout the OptraBot application
"""

from optrabot.exceptions.orderexceptions import OrderException, PlaceOrderException, PrepareOrderException

__all__ = ['OrderException', 'PlaceOrderException', 'PrepareOrderException']


"""
This module contains custom exceptions related to order operations
"""

from optrabot.broker.order import Order


class OrderException(Exception):
    """
    Base class for all order-related exceptions
    """
    def __init__(self, reason: str, order: Order, operation: str = "Unspecific operation", *args):
        """
        Initialize the exception with a message

        Args:
            reason: Reason why the order operation failed
            order: The order that failed to be placed
            args: Additional arguments to pass to the base Exception class
        """
        super().__init__(reason, *args)
        self.reason = reason
        self.operation = operation
        self.order = order

    def __str__(self):
        """
        Return a string representation of the exception
        """
        return f"{self.operation} of order failed: {self.reason}"

class PlaceOrderException(OrderException):
    """
    Exception raised when an order placement fails
    """
    def __init__(self, reason: str, order: Order, *args):
        """
        Initialize the exception with a reason for the failure

        Args:
            reason: The reason why the order placement failed
            order: The order that failed to be placed
            args: Additional arguments to pass to the base Exception class
        """
        super().__init__(reason, order, operation="Place order", *args)

class PrepareOrderException(OrderException):
    """
    Exception raised when preparing an order fails
    """
    def __init__(self, reason: str, order: Order, *args):
        """
        Initialize the exception with a reason for the failure

        Args:
            reason: The reason why the order preparation failed
            order: The order that failed to be prepared
            args: Additional arguments to pass to the base Exception class
        """
        super().__init__(reason, order, operation="Prepare order", *args)


from typing import OrderedDict
from datetime import datetime
import calendar as cal
import pytz

from loguru import logger
class TriggerType:
	External: str = 'external'
	Time: str = 'time'
	Flow: str = 'flow'

class TemplateTrigger:
	_trigger_time: datetime = None
	type: TriggerType = None
	_weekdays: list = None
	_blackout_days: list = None

	def __init__(self, triggerdata: OrderedDict) -> None:
		type = triggerdata['type']
		assert type in [TriggerType.External, TriggerType.Time, TriggerType.Flow]
		self.type = type
		self.value = triggerdata['value']
		assert self.value != '' and self.value != None
		
		if self.type == TriggerType.Time:
			# Check the time 
			try:
				self._trigger_time = self.parseTimeWithTimezone(self.value).astimezone(pytz.utc)
			except ValueError as valError:
				logger.error(f'Invalid time format for parameter "value"! Expecting HH:MM <Timezone>')
				self._trigger_time = None

			# Extract weekdays from the trigger data
			try:
				weekdays = triggerdata['weekdays']
				self._weekdays = self.extractWeekdays(weekdays)
			except KeyError:
				pass
			pass

		try:
			blackout_days = triggerdata['blackout_days']
			self._blackout_days = self.extract_blackout_days(blackout_days)
		except KeyError as keyError:
			pass

		try:
			timefrom = triggerdata['timefrom']
			self.fromTimeUTC = self.parseTimeWithTimezone(timefrom).astimezone(pytz.utc)
		except KeyError:
			self.fromTimeUTC = None
		except ValueError as valError:
			logger.error(f'Invalid time format for parameter "timefrom"!')

		try:
			timeto = triggerdata['timeto']
			self.toTimeUTC = self.parseTimeWithTimezone(timeto).astimezone(pytz.utc)
		except KeyError:
			self.toTimeUTC = None
		except ValueError as valError:
			logger.error(f'Invalid time format for parameter "timeto"!')

		try:
			excludeTimeFrom = triggerdata['excludetimefrom']
			self.excludeFromTimeUTC = self.parseTimeWithTimezone(excludeTimeFrom).astimezone(pytz.utc)
		except KeyError:
			self.excludeFromTimeUTC = None
		except ValueError as valError:
			logger.error(f'Invalid time format for parameter "excludetimefrom"!')

		try:
			excludeTimeTo = triggerdata['excludetimeto']
			self.excludeToTimeUTC = self.parseTimeWithTimezone(excludeTimeTo).astimezone(pytz.utc)
		except KeyError:
			self.excludeToTimeUTC = None
		except ValueError as valError:
			logger.error(f'Invalid time format for parameter "excludetimeto"!')
		
	def extract_blackout_days(self, blackout_days: list) -> list:
		"""
		Extracts the blackout days from the yaml dictionary and stores them in a list.
		"""
		extracted_blackout_days = []
		for value in blackout_days:
			try:
				date = datetime.strptime(value, '%Y-%m-%d').date()
				extracted_blackout_days.append(date)
			except ValueError as valError:
				logger.error(f'Invalid date format "{value}" for parameter "blackout_days"! Expecting YYYY-MM-DD')
				continue

		return extracted_blackout_days if len(extracted_blackout_days) > 0 else None

	def extractWeekdays(self, weekdays: list) -> list:
		"""
		Extracts the weekdays from the yaml dictionary and stores them in a list using the 
		calendar weekday constants.
		"""
		weekdaysList = []
		for value in weekdays:
			uppercase_weekday = value.upper()
			match uppercase_weekday:
				case 'MONDAY':
					weekdaysList.append(cal.MONDAY)
				case 'TUESDAY':
					weekdaysList.append(cal.TUESDAY)
				case 'WEDNESDAY':
					weekdaysList.append(cal.WEDNESDAY)
				case 'THURSDAY':
					weekdaysList.append(cal.THURSDAY)
				case 'FRIDAY':
					weekdaysList.append(cal.FRIDAY)
				case _:
					logger.error(f'Invalid weekday "{value}" in weekdays list! Expecting MONDAY, TUESDAY, WEDNESDAY, THURSDAY, FRIDAY')

		return weekdaysList if len(weekdaysList) > 0 else None

	def parseTimeWithTimezone(self, timeStr: str) -> datetime.time:
		""" 
		Parses a time string like 11:00 EST to a datetime.time object
		"""
		if timeStr == None:
			return None
		
		parts = timeStr.split(' ')
		try:
			now = datetime.now()
			naive_time = datetime.strptime(parts[0], '%H:%M').replace(year=now.year, month=now.month, day=now.day)
			if len(parts) > 1:
				tzstr = parts[1]
			else:
				tzstr = 'EST'
			tzstr = "US/Eastern" if tzstr == 'EST' else tzstr
			tz = pytz.timezone(tzstr)
			localized_time = tz.localize(naive_time)
			return localized_time
		except ValueError and AttributeError as error:
			logger.error(f'Invalid time format: {timeStr} - Expecting HH:MM <Timezone>')
			return None

	def toDict(self):
		""" Returns a dictionary representation of the Trigger which is used for
		the config file.
		"""
		returnDict = {'type': self.type, 'value': self.value}
		eastern = pytz.timezone('US/Eastern')
		if self.fromTimeUTC != None:
			returnDict['timefrom'] = self.fromTimeUTC.astimezone(eastern).strftime('%H:%M %Z')
		if self.toTimeUTC != None:
			returnDict['timeto'] = self.toTimeUTC.astimezone(eastern).strftime('%H:%M %Z')
		if self.excludeFromTimeUTC != None:
			returnDict['excludetimefrom'] = self.excludeFromTimeUTC.astimezone(eastern).strftime('%H:%M %Z')
		if self.excludeToTimeUTC != None:
			returnDict['excludetimeto'] = self.excludeToTimeUTC.astimezone(eastern).strftime('%H:%M %Z')	
		return returnDict


import copy
from datetime import datetime
from typing import List
from loguru import logger
from optrabot.broker.order import Leg, OptionRight, Order, OrderAction, OrderType, PriceEffect
from optrabot.optionhelper import OptionHelper
from optrabot.signaldata import SignalData
from optrabot.tradetemplate.processor.templateprocessorbase import TemplateProcessorBase
from optrabot.tradetemplate.templatefactory import LongPut, Template
from optrabot.managedtrade import ManagedTrade

class LongPutProcessor(TemplateProcessorBase):
	def __init__(self, template: Template):
		"""
		Initializes the Long Put processor with the given template
		"""
		super().__init__(template)

	def composeEntryOrder(self, signalData: SignalData = None):
		"""
		Composes the entry order for the Long Call template
		"""
		super().composeEntryOrder(signalData)
		longPutTemplate :LongPut = self._template
		longStrike = None
		# Long Strike Determination
		if signalData and signalData.strike > 0:
			longStrike = signalData.strike
		else:
			longStrikeData = longPutTemplate.getLongStrikeData()
			if not longStrikeData:
				raise ValueError('Configuration for Long Strike is missing in template!')
			if longStrikeData.offset:
				logger.debug(f'Using Long Strike Offset: {longStrikeData.offset}')
				longStrike = OptionHelper.roundToStrikePrice(signalData.close + longStrikeData.offset)
			if longStrikeData.delta:
				logger.debug(f'Using Long Strike Delta: {longStrikeData.delta}')
				try:
					longStrike = self.get_short_strike_from_delta("SPX", OptionRight.PUT, longStrikeData.delta)
				except ValueError as value_error:
					logger.error(f'Error while determining Short Strike by Delta: {value_error}')
					return None
				
			if longStrike == None:
				raise ValueError('Long Strike could not be determined!')
			
			logger.debug(f'Using Long Strike: {longStrike}')

			# Now create entry order for the long call
			legs: List[Leg] = []
			legs.append(Leg(action=OrderAction.BUY, strike=longStrike, symbol=self._template.symbol, right=OptionRight.PUT, expiration=self._determined_expiration_date, quantity=1))
			entryOrder = Order(symbol=self._template.symbol, legs=legs, action=OrderAction.BUY_TO_OPEN, quantity=self._template.amount, type=OrderType.LIMIT, price_effect=PriceEffect.DEBIT)
		return entryOrder
	
	def composeTakeProfitOrder(self, managedTrade: ManagedTrade, fillPrice: float) -> Order:
		"""
		Composes the take profit order based on the template and the given fill price
		"""
		super().composeTakeProfitOrder(managedTrade, fillPrice)
		logger.debug('Creating take profit order for template {}', self._template.name)
		takeProfitPrice = self._template.calculateTakeProfitPrice(fillPrice)
		logger.debug(f'Calculated take profit price: {takeProfitPrice}')
		tp_order_legs = copy.deepcopy(managedTrade.entryOrder.legs)
		self.invert_leg_actions(tp_order_legs)
		takeProfitOrder = Order(symbol=self._template.symbol, legs=tp_order_legs, action=OrderAction.SELL_TO_CLOSE, quantity=self._template.amount, type=OrderType.LIMIT, price=takeProfitPrice)
		return takeProfitOrder

	def composeStopLossOrder(self, managedTrade: ManagedTrade, fillPrice: float) -> Order:
		"""
		Composes the stop loss order based on the template and the given fill price
		"""
		super().composeStopLossOrder(managedTrade, fillPrice)
		logger.debug('Creating stop loss order for template {}', self._template.name)
		stopLossPrice = self._template.calculateStopLossPrice(fillPrice)
		logger.debug(f'Calculated stop loss price: {stopLossPrice}')
		sl_order_legs = copy.deepcopy(managedTrade.entryOrder.legs)
		self.invert_leg_actions(sl_order_legs)
		stopLossOrder = Order(symbol=self._template.symbol, legs=sl_order_legs, action=OrderAction.SELL_TO_CLOSE, quantity=self._template.amount, type=OrderType.STOP, price=stopLossPrice)
		return stopLossOrder



from typing import List
from optrabot.deltaadjuster import DeltaAdjuster
from optrabot.models import Trade
from optrabot.tradestatus import TradeStatus
from optrabot.tradetemplate.templatefactory import Template
from optrabot.broker.order import Execution, Leg, OptionRight, Order, OrderAction
from optrabot.stoplossadjuster import StopLossAdjuster
import copy

class ManagedTrade:
	"""
	ManagedTrade is representing a trade which is currently managed by the TradeManager.
	"""
	adjustment_orders: List[Order]

	def __init__(self, trade: Trade, template: Template, entryOrder: Order, account: str = ''): 
		self.trade = trade
		self.entryOrder = entryOrder
		self.template = template
		self.account = account
		self.takeProfitOrder: Order = None
		self.stopLossOrder: Order = None
		self.closing_order: Order = None
		self.status = TradeStatus.NEW
		self.realizedPNL = 0.0
		self.transactions = []
		self.expired = False
		self.entry_price = None					# Holds the entry price for the trade
		self.current_price: float = None		# Holds the current price of the trade
		self.current_delta: float = None		# Holds the current delta of the position
		self.current_legs: List[Leg] = []		# Holds the current legs of the trade (filled after entry order is excuted)
		self.stoploss_adjusters: List[StopLossAdjuster] = []
		self.delta_adjusters: List[DeltaAdjuster] = []
		self.long_legs_removed = False			# will be set to true for credit_trades if the long legs are no longer available
		self.adjustment_orders = []				# Holds orders which are created by delta adjusters
		self.entry_adjustment_count = 0			# Tracks the number of entry order price adjustments

	def isActive(self) -> bool:
		"""
		Returns True if the trade is active
		"""
		return self.status == TradeStatus.OPEN
	
	def setup_stoploss_adjusters(self):
		""" 
		Copies the stop loss adjusters from the template to the managed trade and sets the
		base price for earch of the adjusters
		"""
		for adjuster in self.template.get_stoploss_adjusters():
			adjuster_copy = copy.copy(adjuster)
			adjuster_copy.setBasePrice(self.entry_price)
			self.stoploss_adjusters.append(adjuster_copy)

	def setup_delta_adjusters(self):
		"""
		Copies the delta adjusters from the template to the managed trade
		"""
		for adjuster in self.template.get_delta_adjusters():
			adjuster_copy = copy.copy(adjuster)
			self.delta_adjusters.append(adjuster_copy)

	def update_current_legs(self, adjustment_order: Order):
		"""
		Updates the current legs of the trade from the execution of the adjustment order.
		"""
		for leg in adjustment_order.legs:
			opposite_action = OrderAction.SELL if leg.action == OrderAction.BUY else OrderAction.BUY
			existing_leg = next((l for l in self.current_legs if l.strike == leg.strike and l.right == leg.right and l.action != opposite_action), None)
			if existing_leg:
				# If an existing leg was found, remove it from the current legs, because it has been closed
				self.current_legs.remove(existing_leg)
			else:
				# Otherwise it is a new leg added by execution of the adjustment order
				self.current_legs.append(copy.copy(leg))